# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators for reading various types of files from disk."""

import ast
import datetime
import functools
import glob
import logging
import warnings
from pathlib import Path

import iris
import iris.coords
import iris.cube
import iris.util
import numpy as np

from CSET._common import iter_maybe
from CSET.operators._stash_to_lfric import STASH_TO_LFRIC


class NoDataWarning(UserWarning):
    """Warning that no data has been loaded."""


def read_cube(
    loadpath: list[str] | str,
    constraint: iris.Constraint = None,
    filename_pattern: str = "*",
    **kwargs,
) -> iris.cube.Cube:
    """Read a single cube from files.

    Read operator that takes a path string (can include wildcards), and uses
    iris to load the cube matching the constraint.

    If the loaded data is split across multiple files, a filename_pattern can be
    specified to select the read files using Unix shell-style wildcards. In this
    case the loadpath should point to the directory containing the data.

    Ensemble data can also be loaded. If it has a realization coordinate
    already, it will be directly used. If not, it will have its member number
    guessed from the filename, based on one of several common patterns. For
    example the pattern *emXX*, where XX is the realization.

    Deterministic data will be loaded with a realization of 0, allowing it to be
    processed in the same way as ensemble data.

    Arguments
    ---------
    loadpath: str | list[str]
        Path to where .pp/.nc files are located
    constraint: iris.Constraint | iris.ConstraintCombination, optional
        Constraints to filter data by. Defaults to unconstrained.
    filename_pattern: str, optional
        Unix shell-style glob pattern to match filenames to. Defaults to "*"

    Returns
    -------
    cubes: iris.cube.Cube
        Cube loaded

    Raises
    ------
    FileNotFoundError
        If the provided path does not exist
    ValueError
        If the constraint doesn't produce a single cube.
    """
    cubes = read_cubes(loadpath, constraint, filename_pattern)
    # Check filtered cubes is a CubeList containing one cube.
    if len(cubes) == 1:
        return cubes[0]
    else:
        # Log cube details so you can see why they are not merging.
        logging.debug("Non-merging cubes:\n%s", "\n".join(str(cube) for cube in cubes))
        raise ValueError(
            f"Constraint doesn't produce single cube. {constraint}\n{cubes}"
        )


def read_cubes(
    loadpath: list[str] | str,
    constraint: iris.Constraint = None,
    filename_pattern: str = "*",
    **kwargs,
) -> iris.cube.CubeList:
    """Read cubes from files.

    Read operator that takes a path string (can include wildcards), and uses
    iris to load the minimal set of cubes matching the constraint.

    If the loaded data is split across multiple files, a filename_pattern can be
    specified to select the read files using Unix shell-style wildcards. In this
    case the loadpath should point to the directory containing the data.

    Ensemble data can also be loaded. If it has a realization coordinate
    already, it will be directly used. If not, it will have its member number
    guessed from the filename, based on one of several common patterns. For
    example the pattern *emXX*, where XX is the realization.

    Deterministic data will be loaded with a realization of 0, allowing it to be
    processed in the same way as ensemble data.

    Data output by XIOS (such as LFRic) has its per-file metadata removed so
    that the cubes merge across files.

    Arguments
    ---------
    loadpath: str | list[str]
        Path to where .pp/.nc files are located. Can include globs.
    constraint: iris.Constraint | iris.ConstraintCombination, optional
        Constraints to filter data by. Defaults to unconstrained.
    filename_pattern: str, optional
        Unix shell-style glob pattern to match filenames to. Defaults to "*"

    Returns
    -------
    cubes: iris.cube.CubeList
        Cubes loaded after being merged and concatenated.

    Raises
    ------
    FileNotFoundError
        If the provided path does not exist
    """
    logging.debug("Constraint: %s", constraint)

    def load_from_paths(
        paths: list[str], is_ensemble: bool, is_base: bool
    ) -> iris.cube.CubeList:
        # Skip if no paths given, mainly for when we only have a base path.
        if not paths:
            return iris.cube.CubeList([])
        input_files = _check_input_files(paths, filename_pattern)
        callback = _create_callback(is_ensemble=is_ensemble, is_base=is_base)
        # If unset, a constraint of None lets everything be loaded.
        return iris.load(input_files, constraint, callback=callback)

    # Get lists of paths.
    paths = iter_maybe(loadpath)
    base_path = paths[:1]
    # If len(paths) < 2, this just returns an empty list, rather than an error.
    other_paths = paths[1:]
    del paths

    # Load the data, then reload with correct handling if it is ensemble data.
    cubes = iris.cube.CubeList([])

    # Split out first input file definition here and mark as base model.
    cubes.extend(load_from_paths(base_path, is_ensemble=False, is_base=True))
    cubes.extend(load_from_paths(other_paths, is_ensemble=False, is_base=False))

    # Reload all data with ensemble handling if needed.
    if _is_ensemble(cubes):
        cubes = iris.cube.CubeList()
        cubes.extend(load_from_paths(base_path, is_ensemble=True, is_base=True))
        cubes.extend(load_from_paths(other_paths, is_ensemble=True, is_base=False))

    # Merge and concatenate cubes now metadata has been fixed.
    cubes = cubes.merge()
    cubes = cubes.concatenate()

    # Ensure dimension coordinates are bounded.
    for cube in cubes:
        for dim_coord in cube.coords(dim_coords=True):
            # Iris can't guess the bounds of a scalar coordinate.
            if not dim_coord.has_bounds() and dim_coord.shape[0] > 1:
                dim_coord.guess_bounds()

    logging.debug("Loaded cubes: %s", cubes)
    if len(cubes) == 0:
        warnings.warn(
            "No cubes loaded, check your constraints!", NoDataWarning, stacklevel=2
        )
    return cubes


def _is_ensemble(cubelist: iris.cube.CubeList) -> bool:
    """Test if a CubeList is likely to be ensemble data.

    If cubes either have a realization dimension, or there are multiple files
    for the same time-step, we can assume it is ensemble data.
    """
    unique_cubes = set()
    for cube in cubelist:
        # Ignore realization of 0, as that is given to deterministic data.
        if cube.coords("realization") and any(cube.coord("realization").points != 0):
            return True
        # Compare XML representation of cube structure check for duplicates.
        cube_content = cube.xml()
        if cube_content in unique_cubes:
            logging.info("Ensemble data loaded.")
            logging.debug("Cube Contents: %s", unique_cubes)
            return True
        else:
            unique_cubes.add(cube_content)
    logging.info("Deterministic data loaded.")
    logging.debug("Cube Contents: %s", unique_cubes)
    return False


def _create_callback(is_ensemble: bool, is_base: bool) -> callable:
    """Compose together the needed callbacks into a single function."""

    # TODO: Fix coordinate name to enable full level and half level merging.
    def callback(cube: iris.cube.Cube, field, filename: str):
        if is_base:
            _mark_as_comparison_base_callback(cube, field, filename)
        if is_ensemble:
            _ensemble_callback(cube, field, filename)
        else:
            _deterministic_callback(cube, field, filename)
        _um_normalise_callback(cube, field, filename)
        _lfric_normalise_callback(cube, field, filename)
        _lfric_time_coord_fix_callback(cube, field, filename)
        _longitude_fix_callback(cube, field, filename)
        _fix_spatial_coord_name_callback(cube)
        _lfric_time_callback(cube)
        _fix_pressure_coord_callback(cube)
        _fix_um_radtime_prehour(cube)
        _fix_um_radtime_posthour(cube)
        _fix_lfric_longnames(cube)
        _fix_um_lightning(cube)
        _fix_lfric_radtime_prehour(cube)

    return callback


def _mark_as_comparison_base_callback(cube, field, filename):
    """Add an attribute to the cube to mark it as the base for comparisons."""
    # Use 1 to indicate True, as booleans can't be saved in NetCDF attributes.
    cube.attributes["cset_comparison_base"] = 1


def _ensemble_callback(cube, field, filename):
    """Add a realization coordinate to a cube.

    Uses the filename to add an ensemble member ('realization') to each cube.
    Assumes data is formatted enuk_um_0XX/enukaa_pd0HH.pp where XX is the
    ensemble member.

    Arguments
    ---------
    cube: Cube
        ensemble member cube
    field
        Raw data variable, unused.
    filename: str
        filename of ensemble member data
    """
    if not cube.coords("realization"):
        if "em" in filename:
            # Assuming format is *emXX*
            loc = filename.find("em") + 2
            member = np.int32(filename[loc : loc + 2])
        else:
            # Assuming raw fields files format is enuk_um_0XX/enukaa_pd0HH
            member = np.int32(filename[-15:-13])

        cube.add_aux_coord(iris.coords.AuxCoord(member, standard_name="realization"))


def _deterministic_callback(cube, field, filename):
    """Give deterministic cubes a realization of 0.

    This means they can be handled in the same way as ensembles through the rest
    of the code.
    """
    # Only add if realization coordinate does not exist.
    if not cube.coords("realization"):
        cube.add_aux_coord(
            iris.coords.AuxCoord(np.int32(0), standard_name="realization", units="1")
        )


@functools.lru_cache(None)
def _warn_once(msg):
    """Print a warning message, skipping recent duplicates."""
    logging.warning(msg)


def _um_normalise_callback(cube: iris.cube.Cube, field, filename):
    """Normalise UM STASH variable long names to LFRic variable names.

    Note standard names will remain associated with cubes where different.
    Long name will be used consistently in output filename and titles.
    """
    # Convert STASH to LFRic variable name
    if "STASH" in cube.attributes:
        stash = cube.attributes["STASH"]
        try:
            (name, grid) = STASH_TO_LFRIC[str(stash)]
            cube.long_name = name
        except KeyError:
            # Don't change cubes with unknown stash codes.
            _warn_once(
                f"Unknown STASH code: {stash}. Please check file stash_to_lfric.py to update."
            )


def _lfric_normalise_callback(cube: iris.cube.Cube, field, filename):
    """Normalise attributes that prevents LFRic cube from merging.

    The uuid and timeStamp relate to the output file, as saved by XIOS, and has
    no relation to the data contained. These attributes are removed.

    The um_stash_source is a list of STASH codes for when an LFRic field maps to
    multiple UM fields, however it can be encoded in any order. This attribute
    is sorted to prevent this. This attribute is only present in LFRic data that
    has been converted to look like UM data.
    """
    # Remove unwanted attributes.
    cube.attributes.pop("timeStamp", None)
    cube.attributes.pop("uuid", None)
    cube.attributes.pop("name", None)

    # Sort STASH code list.
    stash_list = cube.attributes.get("um_stash_source")
    if stash_list:
        # Parse the string as a list, sort, then re-encode as a string.
        cube.attributes["um_stash_source"] = str(sorted(ast.literal_eval(stash_list)))


def _lfric_time_coord_fix_callback(cube: iris.cube.Cube, field, filename):
    """Ensure the time coordinate is a DimCoord rather than an AuxCoord.

    The coordinate is converted and replaced if not. SLAMed LFRic data has this
    issue, though the coordinate satisfies all the properties for a DimCoord.
    Scalar time values are left as AuxCoords.
    """
    # This issue seems to come from iris's handling of NetCDF files where time
    # always ends up as an AuxCoord.
    if cube.coords("time"):
        time_coord = cube.coord("time")
        if (
            not isinstance(time_coord, iris.coords.DimCoord)
            and len(cube.coord_dims(time_coord)) == 1
        ):
            iris.util.promote_aux_coord_to_dim_coord(cube, time_coord)

    # Force single-valued coordinates to be scalar coordinates.
    return iris.util.squeeze(cube)


def _longitude_fix_callback(cube: iris.cube.Cube, field, filename):
    """Check longitude coordinates are in the range -180 deg to 180 deg.

    This is necessary if comparing two models with different conventions --
    for example, models where the prime meridian is defined as 0 deg or
    360 deg. If not in the range -180 deg to 180 deg, we wrap the longitude
    so that it falls in this range.
    """
    import CSET.operators._utils as utils

    try:
        y, x = utils.get_cube_yxcoordname(cube)
    except ValueError:
        # Don't modify non-spatial cubes.
        return cube
    long_coord = cube.coord(x)
    long_points = long_coord.points.copy()
    long_centre = np.median(long_points)
    while long_centre < -180.0:
        long_centre += 360.0
        long_points += 360.0
    while long_centre >= 180.0:
        long_centre -= 360.0
        long_points -= 360.0
    long_coord.points = long_points
    return cube


def _fix_spatial_coord_name_callback(cube: iris.cube.Cube):
    """Check latitude and longitude coordinates name.

    This is necessary as some models define their grid as 'grid_latitude' and 'grid_longitude'
    and this means that recipes will fail - particularly if the user is comparing multiple models
    where the spatial coordinate names differ.
    """
    import CSET.operators._utils as utils

    # Check if cube is spatial.
    if not utils.is_spatialdim(cube):
        # Don't modify non-spatial cubes.
        return cube

    # Get spatial coords.
    y_name, x_name = utils.get_cube_yxcoordname(cube)

    if y_name in ["latitude"] and cube.coord(y_name).units in [
        "degrees",
        "degrees_north",
        "degrees_south",
    ]:
        cube.coord(y_name).rename("grid_latitude")
    if x_name in ["longitude"] and cube.coord(x_name).units in [
        "degrees",
        "degrees_west",
        "degrees_east",
    ]:
        cube.coord(x_name).rename("grid_longitude")


def _fix_pressure_coord_callback(cube: iris.cube.Cube):
    """Rename pressure coordinate to "pressure" if it exists and ensure hPa units.

    This problem was raised because the AIFS model data from ECMWF
    defines the pressure coordinate with the name "pressure_level" rather
    than compliant CF coordinate names.

    Additionally, set the units of pressure to be hPa to be consistent with the UM,
    and approach the coordinates in a unified way.
    """
    for coord in cube.dim_coords:
        if coord.name() in ["pressure_level", "pressure_levels"]:
            coord.rename("pressure")

        if coord.name() == "pressure":
            if str(cube.coord("pressure").units) != "hPa":
                cube.coord("pressure").convert_units("hPa")


def _fix_um_radtime_posthour(cube: iris.cube.Cube):
    """Fix radiation which is output 1 minute past every hour."""
    try:
        if cube.attributes["STASH"] in [
            "m01s01i208",
            "m01s02i205",
            "m01s02i201",
            "m01s01i207",
            "m01s02i207",
            "m01s01i235",
        ]:
            time_coord = cube.coord("time")

            # Convert time points to datetime objects
            time_unit = time_coord.units
            time_points = time_unit.num2date(time_coord.points)

            # Subtract 1 minute from each time point
            new_time_points = np.array(
                [t - datetime.timedelta(minutes=1) for t in time_points]
            )

            # Convert back to numeric values using the original time unit
            new_time_values = time_unit.date2num(new_time_points)

            # Replace the time coordinate with corrected values
            time_coord.points = new_time_values
    except KeyError:
        pass


def _fix_um_radtime_prehour(cube: iris.cube.Cube):
    """Fix radiation which is output 1 minute before every hour."""
    try:
        if cube.attributes["STASH"] == "m01s01i207":
            time_coord = cube.coord("time")

            # Convert time points to datetime objects
            time_unit = time_coord.units
            time_points = time_unit.num2date(time_coord.points)

            # Add 1 minute from each time point
            new_time_points = np.array(
                [t + datetime.timedelta(minutes=1) for t in time_points]
            )

            # Convert back to numeric values using the original time unit
            new_time_values = time_unit.date2num(new_time_points)

            # Replace the time coordinate with corrected values
            time_coord.points = new_time_values
    except KeyError:
        pass


def _fix_lfric_radtime_prehour(cube: iris.cube.Cube):
    """Fix radiation which is output 1 minute before every hour."""
    try:
        # Check LFRic.
        if "time_origin" in cube.coord("time").attributes and cube.name() in [
            "toa_upward_longwave_flux_radiative_timestep",
            "surface_net_longwave_flux_radiative_timestep",
        ]:
            time_coord = cube.coord("time")

            # Convert time points to datetime objects
            time_unit = time_coord.units
            time_points = time_unit.num2date(time_coord.points)

            # Add 1 minute from each time point
            new_time_points = np.array(
                [t + datetime.timedelta(minutes=1) for t in time_points]
            )

            # Convert back to numeric values using the original time unit
            new_time_values = time_unit.date2num(new_time_points)

            # Replace the time coordinate with corrected values
            time_coord.points = new_time_values
    except KeyError:
        pass


def _fix_lfric_longnames(cube: iris.cube.Cube):
    """To fix names where the long name is appropriate."""
    try:
        if str(cube.long_name) == "combined_cloud_amount_maximum_random_overlap":
            cube.rename("combined_cloud_amount_maximum_random_overlap")
    except KeyError:
        pass


def _fix_um_lightning(cube: iris.cube.Cube):
    """To fix the date points in lightning accumulation STASH.

    Lightning (m01s21i104) is being output as a time accumulation in UM,
    over each hour (TAcc1hr), not from the start of the forecast, to be compatible
    with LFRic. So this is a short term solution to remove cell methods (
    as variables are ignored with cell methods for surface plots currently),
    and also adjust the time so that the value is at the end of each hour.
    """
    try:
        if cube.attributes["STASH"] == "m01s21i104":
            cube.cell_methods = []

            time_coord = cube.coord("time")

            # Convert time points to datetime objects
            time_unit = time_coord.units
            time_points = time_unit.num2date(time_coord.points)

            # Subtract 1 minute from each time point
            new_time_points = np.array(
                [t + datetime.timedelta(minutes=30) for t in time_points]
            )

            # Convert back to numeric values using the original time unit
            new_time_values = time_unit.date2num(new_time_points)

            # Replace the time coordinate with corrected values
            time_coord.points = new_time_values
    except KeyError:
        pass


def _lfric_time_callback(cube: iris.cube.Cube):
    """Fix time coordinate metadata if missing dimensions.

    Some model data does not contain forecast_reference_time or forecast_period as
    expected coordinates, and so we cannot aggregate over case studies without this
    metadata. This callback fixes these issues.

    Notes
    -----
    Some parts of the code have been adapted from Paul Earnshaw's scripts.
    """
    # See what coordinates exist in cube
    coord_names = [coord.name() for coord in cube.coords()]

    # Construct forecast_reference time and forecast_period if they dont exist.
    if "forecast_reference_time" not in coord_names:
        if "time" in coord_names:
            tcoord = cube.coord("time")
            if "time_origin" in tcoord.attributes:
                init_time = datetime.datetime.fromisoformat(
                    cube.coord("time").attributes["time_origin"]
                )
                time_unit = cube.coord("time").units
                frt_point = time_unit.date2num(init_time)

                frt_coord = iris.coords.AuxCoord(
                    frt_point,
                    units=tcoord.units,
                    standard_name="forecast_reference_time",
                )
                cube.add_aux_coord(frt_coord)
            else:
                logging.info(
                    "Cannot find forecast_reference_time, but no time_origin to construct it"
                )

        else:
            logging.info(
                "Cannot find forecast_reference_time, but no time axis to construct it"
            )

    # Regenerate coordinates list from cube
    coord_names = [coord.name() for coord in cube.coords()]

    # Construct forecast_period axis (leadtime) if doesn't exist.
    if "forecast_period" not in coord_names:
        if "time" in coord_names and "forecast_reference_time" in coord_names:
            # Get time coordinate and forecast_reference_time
            tcoord = cube.coord("time")
            init = cube.coord("forecast_reference_time")

            # Create array of forecast leadtimes.
            lead_times = tcoord.points - init.points
            if "seconds" in str(tcoord.units):
                units = "seconds"
            elif "hours" in str(tcoord.units):
                units = "hours"
            else:
                raise ValueError(f"Not a recognised base time unit {str(tcoord.units)}")

            # Create new axis object.
            lead_time_coord = iris.coords.AuxCoord(
                lead_times, long_name="forecast_period", units=units
            )

            # Associate lead time dimension with coordinate time.
            time_dim = cube.coord_dims("time")
            cube.add_aux_coord(lead_time_coord, time_dim)

        else:
            logging.info(
                "No time coordinate or forecast_reference_time, so cannot construct forecast_period"
            )


def _check_input_files(input_paths: list[str], filename_pattern: str) -> list[Path]:
    """Get an iterable of files to load, and check that they all exist.

    Arguments
    ---------
    input_paths: list[str]
        List of paths to input files or directories. The path may itself contain
        glob patterns, but unlike in shells it will match directly first.

    filename_pattern: str
        Shell-style glob pattern to match inside an input directory.

    Returns
    -------
    list[Path]
        A list of files to load.

    Raises
    ------
    FileNotFoundError:
        If the provided arguments don't resolve to at least one existing file.
    """
    files = []
    for raw_filename in iter_maybe(input_paths):
        # Match glob-like files first, if they exist.
        raw_path = Path(raw_filename)
        if raw_path.is_file():
            files.append(raw_path)
        else:
            for input_path in glob.glob(str(raw_filename)):
                # Convert string paths into Path objects.
                input_path = Path(input_path)
                # Get the list of files in the directory, or use it directly.
                if input_path.is_dir():
                    logging.debug(
                        "Checking directory '%s' for files matching '%s'",
                        input_path,
                        filename_pattern,
                    )
                    files.extend(input_path.glob(filename_pattern))
                else:
                    files.append(input_path)

    files.sort()
    logging.info("Loading files:\n%s", "\n".join(str(path) for path in files))
    if len(files) == 0:
        raise FileNotFoundError(f"No files found for {input_paths}")
    return files



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test plotting operators."""

import json
from pathlib import Path

import iris.cube
import matplotlib as mpl
import numpy as np
import pytest

from CSET.operators import collapse, plot, read


def test_check_single_cube():
    """Conversion to a single cube, and rejection where not possible."""
    cube = iris.cube.Cube([0.0])
    cubelist = iris.cube.CubeList([cube])
    long_cubelist = iris.cube.CubeList([cube, cube])
    non_cube = 1
    assert plot._check_single_cube(cube) == cube
    assert plot._check_single_cube(cubelist) == cube
    with pytest.raises(TypeError):
        plot._check_single_cube(long_cubelist)
    with pytest.raises(TypeError):
        plot._check_single_cube(non_cube)


def test_load_colorbar_map():
    """Colorbar is loaded correctly."""
    colorbar = plot._load_colorbar_map()
    assert isinstance(colorbar, dict)
    # Check we can find an example definition.
    assert colorbar["temperature_at_screen_level"] == {
        "cmap": "RdYlBu_r",
        "max": 323,
        "min": 223,
    }


def test_load_colorbar_map_override(tmp_path):
    """Colorbar is loaded correctly and overridden by the user definition."""
    # Setup a user provided colorbar override.
    user_definition = {"temperature_at_screen_level": {"max": 1000, "min": 0}}
    user_colorbar_file = tmp_path / "colorbar.json"
    with open(user_colorbar_file, "wt") as fp:
        json.dump(user_definition, fp)

    colorbar = plot._load_colorbar_map(user_colorbar_file)

    assert isinstance(colorbar, dict)
    # Check definition is updated.
    assert colorbar["temperature_at_screen_level"] == {
        "cmap": "RdYlBu_r",
        "max": 1000,
        "min": 0,
    }
    # Check we can still see unchanged definitions.
    assert colorbar["temperature_at_screen_level_difference"] == {
        "cmap": "bwr",
        "max": 2,
        "min": -2,
    }


def test_load_colorbar_map_override_file_not_found(tmp_path):
    """Colorbar overridden by the user definition in non-existent file."""
    user_colorbar_file = tmp_path / "colorbar.json"
    colorbar = plot._load_colorbar_map(user_colorbar_file)
    # Check it still returns the built-in one.
    assert isinstance(colorbar, dict)


def test_colorbar_map_levels(cube, tmp_working_dir):
    """Colorbar definition is found for cube."""
    cmap, levels, norm = plot._colorbar_map_levels(cube)
    assert cmap == mpl.colormaps["RdYlBu_r"]
    assert (levels == np.linspace(223, 323, 51)).all()
    assert norm is None


# Test will fail (but not cause an overall fail) as we can't test using levels
# until there is at least one variable that uses them.
@pytest.mark.xfail(reason="No colorbar currently uses levels")
def test_colorbar_map_levels_def_on_levels(cube, tmp_working_dir):
    """Colorbar definition that uses levels is found for cube."""
    cmap, levels, norm = plot._colorbar_map_levels(cube)
    assert cmap == mpl.colormaps[...]
    assert levels == [1, 2, 3]
    assert norm == ...


def test_colorbar_map_levels_name_fallback(cube, tmp_working_dir):
    """Colorbar definition is found for cube after checking its other names."""
    cube.standard_name = None
    cmap, levels, norm = plot._colorbar_map_levels(cube)
    assert cmap == mpl.colormaps["RdYlBu_r"]
    assert (levels == np.linspace(223, 323, 51)).all()
    assert norm is None


def test_colorbar_map_levels_unknown_variable_fallback(cube, tmp_working_dir):
    """Colorbar definition doesn't exist for cube."""
    cube.standard_name = None
    cube.long_name = None
    cube.var_name = "unknown"
    cmap, levels, norm = plot._colorbar_map_levels(cube)
    assert cmap == mpl.colormaps["viridis"]
    assert levels is None
    assert norm is None


def test_colorbar_map_levels_pressure_level(transect_source_cube, tmp_working_dir):
    """Pressure level specific colorbar definition is picked up."""
    cube_250hPa = transect_source_cube.extract(iris.Constraint(pressure=250))
    cmap, levels, norm = plot._colorbar_map_levels(cube_250hPa)
    assert cmap == mpl.colormaps["RdYlBu_r"]
    assert (levels == np.linspace(200, 240, 51)).all()
    assert norm is None


def test_spatial_contour_plot(cube, tmp_working_dir):
    """Plot spatial contour plot of instant air temp."""
    # Remove realization coord to increase coverage, and as its not needed.
    cube.remove_coord("realization")
    cube_2d = cube.slices_over("time").next()
    plot.spatial_contour_plot(cube_2d, filename="plot")
    assert Path("plot_462147.0.png").is_file()


def test_contour_plot_sequence(cube, tmp_working_dir):
    """Plot sequence of contour plots."""
    plot.spatial_contour_plot(cube, sequence_coordinate="time")
    assert Path("untitled_462147.0.png").is_file()
    assert Path("untitled_462148.0.png").is_file()
    assert Path("untitled_462149.0.png").is_file()


def test_postage_stamp_contour_plot(monkeypatch, tmp_path):
    """Plot postage stamp plots of ensemble data."""
    ensemble_cube = read.read_cube("tests/test_data/exeter_em*.nc")
    # Get a single time step.
    ensemble_cube_3d = next(ensemble_cube.slices_over("time"))
    monkeypatch.chdir(tmp_path)
    plot.spatial_contour_plot(ensemble_cube_3d)
    assert Path("untitled_463858.0.png").is_file()


def test_postage_stamp_contour_plot_sequence_coord_check(cube, tmp_working_dir):
    """Check error when cube has no time coordinate."""
    # What does this even physically mean? No data?
    cube.remove_coord("time")
    with pytest.raises(ValueError):
        plot.spatial_contour_plot(cube)


def test_spatial_pcolormesh_plot(cube, tmp_working_dir):
    """Plot spatial pcolormesh plot of instant air temp."""
    # Remove realization coord to increase coverage, and as its not needed.
    cube.remove_coord("realization")
    cube_2d = cube.slices_over("time").next()
    plot.spatial_pcolormesh_plot(cube_2d, filename="plot")
    assert Path("plot_462147.0.png").is_file()


def test_pcolormesh_plot_sequence(cube, tmp_working_dir):
    """Plot sequence of pcolormesh plots."""
    plot.spatial_pcolormesh_plot(cube, sequence_coordinate="time")
    assert Path("untitled_462147.0.png").is_file()
    assert Path("untitled_462148.0.png").is_file()
    assert Path("untitled_462149.0.png").is_file()


def test_postage_stamp_pcolormesh_plot(monkeypatch, tmp_path):
    """Plot postage stamp plots of ensemble data."""
    ensemble_cube = read.read_cube("tests/test_data/exeter_em*.nc")
    # Get a single time step.
    ensemble_cube_3d = next(ensemble_cube.slices_over("time"))
    monkeypatch.chdir(tmp_path)
    plot.spatial_pcolormesh_plot(ensemble_cube_3d)
    assert Path("untitled_463858.0.png").is_file()


def test_postage_stamp_pcolormesh_plot_sequence_coord_check(cube, tmp_working_dir):
    """Check error when cube has no time coordinate."""
    # What does this even physically mean? No data?
    cube.remove_coord("time")
    with pytest.raises(ValueError):
        plot.spatial_pcolormesh_plot(cube)


def test_plot_line_series(cube, tmp_working_dir):
    """Save a line series plot."""
    cube = collapse.collapse(cube, ["grid_latitude", "grid_longitude"], "MEAN")
    plot.plot_line_series(cube)
    assert Path("untitled.png").is_file()


def test_plot_line_series_with_filename(cube, tmp_working_dir):
    """Save a line series plot with specific filename and series coordinate."""
    cube = collapse.collapse(cube, ["time", "grid_longitude"], "MEAN")
    plot.plot_line_series(
        cube, filename="latitude_average.ext", series_coordinate="grid_latitude"
    )
    assert Path("latitude_average.png").is_file()


def test_plot_line_series_no_series_coordinate(tmp_working_dir):
    """Error when cube is missing series coordinate (time)."""
    cube = iris.cube.Cube([], var_name="nothing")
    with pytest.raises(ValueError):
        plot.plot_line_series(cube)


def test_plot_line_series_too_many_dimensions(cube, tmp_working_dir):
    """Error when cube has more than one dimension."""
    with pytest.raises(ValueError):
        plot.plot_line_series(cube)


def test_plot_vertical_line_series(vertical_profile_cube, tmp_working_dir):
    """Save a vertical line series plot."""
    plot.plot_vertical_line_series(
        vertical_profile_cube, series_coordinate="pressure", sequence_coordinate="time"
    )
    assert Path("untitled_473718.0.png").is_file()
    assert Path("untitled_473721.0.png").is_file()


def test_plot_vertical_line_series_with_filename(
    vertical_profile_cube, tmp_working_dir
):
    """Save a vertical line series plot with specific filename.

    The given filename does not haven extension to test that too.
    """
    plot.plot_vertical_line_series(
        vertical_profile_cube,
        filename="Test",
        series_coordinate="pressure",
        sequence_coordinate="time",
    )
    assert Path("Test_473718.0.png").is_file()
    assert Path("Test_473721.0.png").is_file()


def test_plot_vertical_line_series_no_series_coordinate(
    vertical_profile_cube, tmp_working_dir
):
    """Error when cube is missing series coordinate (pressure)."""
    vertical_profile_cube.remove_coord("pressure")
    with pytest.raises(ValueError, match="Cube must have a pressure coordinate."):
        plot.plot_vertical_line_series(
            vertical_profile_cube, series_coordinate="pressure"
        )


def test_plot_vertical_line_series_no_sequence_coordinate(
    vertical_profile_cube, tmp_working_dir
):
    """Error when cube is missing sequence coordinate (time)."""
    vertical_profile_cube.remove_coord("time")
    with pytest.raises(ValueError, match="Cube must have a time coordinate."):
        plot.plot_vertical_line_series(
            vertical_profile_cube, series_coordinate="pressure"
        )


def test_plot_vertical_line_series_too_many_dimensions(cube, tmp_working_dir):
    """Error when cube has more than one dimension."""
    with pytest.raises(
        ValueError, match="Cube must have a model_level_number coordinate."
    ):
        plot.plot_vertical_line_series(cube)


def test_plot_histogram_no_sequence_coordinate(histogram_cube, tmp_working_dir):
    """Error when cube is missing sequence coordinate (time)."""
    histogram_cube.remove_coord("time")
    with pytest.raises(ValueError, match="Cube must have a time coordinate."):
        plot.plot_histogram_series(histogram_cube, series_coordinate="pressure")


def test_plot_histogram_with_filename(histogram_cube, tmp_working_dir):
    """Plot sequence of contour plots."""
    plot.plot_histogram_series(
        histogram_cube, filename="test", sequence_coordinate="time"
    )
    assert Path("test_473718.0.png").is_file()
    assert Path("test_473721.0.png").is_file()


def test_plot_and_save_postage_stamp_histogram_series(histogram_cube, tmp_working_dir):
    """Test plotting a postage stamp histogram."""
    plot._plot_and_save_postage_stamp_histogram_series(
        cube=histogram_cube,
        filename="test.png",
        title="Test",
        stamp_coordinate="realization",
        vmin=250,
        vmax=350,
        histtype="step",
    )
    assert Path("test.png").is_file()


def test_plot_and_save_postage_stamps_in_single_plot_histogram_series(
    histogram_cube, tmp_working_dir
):
    """Test plotting a multiline histogram for multiple ensemble members."""
    plot._plot_and_save_postage_stamps_in_single_plot_histogram_series(
        cube=histogram_cube,
        filename="test.png",
        title="Test",
        stamp_coordinate="realization",
        vmin=250,
        vmax=350,
        histtype="step",
    )
    assert Path("test.png").is_file()


def test_scatter_plot(cube, vertical_profile_cube, tmp_working_dir):
    """Save a scatter plot."""
    cube_y = collapse.collapse(cube, ["time", "grid_longitude"], "MEAN")[0:4]
    cube_x = collapse.collapse(vertical_profile_cube, ["time"], "MEAN")[0:4]
    plot.scatter_plot(
        cube_y,
        cube_x,
    )
    assert Path("untitled.png").is_file()


def test_scatter_plot_with_filename(cube, vertical_profile_cube, tmp_working_dir):
    """Save a scatter plot with a file name."""
    cube_y = collapse.collapse(cube, ["time", "grid_longitude"], "MEAN")[0:4]
    cube_x = collapse.collapse(vertical_profile_cube, ["time"], "MEAN")[0:4]
    plot.scatter_plot(
        cube_y,
        cube_x,
        filename="scatter_plot.ext",
    )
    assert Path("scatter_plot.png").is_file()


def test_scatter_plot_no_one_to_one_line(cube, vertical_profile_cube, tmp_working_dir):
    """Save a scatter plot without a one-to-one line."""
    cube_y = collapse.collapse(cube, ["time", "grid_longitude"], "MEAN")[0:4]
    cube_x = collapse.collapse(vertical_profile_cube, ["time"], "MEAN")[0:4]
    plot.scatter_plot(
        cube_y,
        cube_x,
        one_to_one=False,
    )
    assert Path("untitled.png").is_file()


def test_scatter_plot_too_many_x_dimensions(
    cube, vertical_profile_cube, tmp_working_dir
):
    """Error when cube_x has more than one dimension."""
    cube_y = collapse.collapse(cube, ["time", "grid_longitude"], "MEAN")[0:4]
    cube_x = vertical_profile_cube.copy()
    with pytest.raises(ValueError):
        plot.scatter_plot(cube_x, cube_y)


def test_scatter_plot_too_many_y_dimensions(
    cube, vertical_profile_cube, tmp_working_dir
):
    """Error when cube_y has more than one dimension."""
    cube_y = cube.copy()
    cube_x = collapse.collapse(vertical_profile_cube, ["time"], "MEAN")[0:4]
    with pytest.raises(ValueError):
        plot.scatter_plot(cube_x, cube_y)


def test_get_plot_resolution(tmp_working_dir):
    """Test getting the plot resolution."""
    with open("meta.json", "wt", encoding="UTF-8") as fp:
        fp.write('{"plot_resolution": 72}')
    resolution = plot._get_plot_resolution()
    assert resolution == 72


def test_get_plot_resolution_unset(tmp_working_dir):
    """Test getting the default plot resolution when unset."""
    resolution = plot._get_plot_resolution()
    assert resolution == 100


def test_invalid_plotting_method_spatial_plot(cube, tmp_working_dir):
    """Test plotting a spatial plot with an invalid method."""
    with pytest.raises(ValueError, match="Unknown plotting method"):
        plot._plot_and_save_spatial_plot(cube, "filename", "title", "invalid")


def test_invalid_plotting_method_postage_stamp_spatial_plot(cube, tmp_working_dir):
    """Test plotting a postage stamp spatial plot with an invalid method."""
    with pytest.raises(ValueError, match="Unknown plotting method"):
        plot._plot_and_save_postage_stamp_spatial_plot(
            cube, "filename", "realization", "title", "invalid"
        )



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators to regrid cubes."""

import warnings

import iris
import iris.cube
import numpy as np

from CSET._common import iter_maybe
from CSET.operators._utils import get_cube_yxcoordname


class BoundaryWarning(UserWarning):
    """Selected gridpoint is close to the domain edge.

    In many cases gridpoints near the domain boundary contain non-physical
    values, so caution is advised when interpreting them.
    """


def regrid_onto_cube(
    toregrid: iris.cube.Cube | iris.cube.CubeList,
    target: iris.cube.Cube,
    method: str,
    **kwargs,
) -> iris.cube.Cube | iris.cube.CubeList:
    """Regrid a cube or CubeList, projecting onto a target cube.

    All cubes must have at least 2 spatial (map) dimensions.

    Arguments
    ----------
    toregrid: iris.cube | iris.cube.CubeList
        An iris Cube of data to regrid, or multiple cubes to regrid in a
        CubeList. A minimum requirement is that the cube(s) need to be 2D with a
        latitude, longitude coordinates.
    target: Cube
        An iris cube of the data to regrid onto. It needs to be 2D with a
        latitude, longitude coordinate.
    method: str
        Method used to regrid onto, etc. Linear will use iris.analysis.Linear()

    Returns
    -------
    iris.cube | iris.cube.CubeList
        An iris cube of the data that has been regridded, or a CubeList of the
        cubes that have been regridded in the same order they were passed in
        toregrid.

    Raises
    ------
    ValueError
        If a unique x/y coordinate cannot be found
    NotImplementedError
        If the cubes grid, or the method for regridding, is not yet supported.

    Notes
    -----
    Currently rectlinear grids (uniform) are supported.
    """
    # To store regridded cubes.
    regridded_cubes = iris.cube.CubeList()

    # Iterate over all cubes and regrid.
    for cube in iter_maybe(toregrid):
        # Get y,x coord names
        y_coord, x_coord = get_cube_yxcoordname(cube)

        # List of supported grids - check if it is compatible
        supported_grids = (iris.coord_systems.GeogCS, iris.coord_systems.RotatedGeogCS)
        if not isinstance(cube.coord(x_coord).coord_system, supported_grids):
            raise NotImplementedError(
                f"Does not currently support {cube.coord(x_coord).coord_system} coordinate system"
            )
        if not isinstance(cube.coord(y_coord).coord_system, supported_grids):
            raise NotImplementedError(
                f"Does not currently support {cube.coord(y_coord).coord_system} coordinate system"
            )

        regrid_method = getattr(iris.analysis, method, None)
        if callable(regrid_method):
            regridded_cubes.append(cube.regrid(target, regrid_method()))
        else:
            raise NotImplementedError(
                f"Does not currently support {method} regrid method"
            )

    # Preserve returning a cube if only a cube has been supplied to regrid.
    if len(regridded_cubes) == 1:
        return regridded_cubes[0]
    else:
        return regridded_cubes


def regrid_onto_xyspacing(
    toregrid: iris.cube.Cube | iris.cube.CubeList,
    xspacing: float,
    yspacing: float,
    method: str,
    **kwargs,
) -> iris.cube.Cube | iris.cube.CubeList:
    """Regrid cube or cubelist onto a set x,y spacing.

    Regrid cube(s) using specified x,y spacing, which is performed linearly.

    Parameters
    ----------
    toregrid: iris.cube | iris.cube.CubeList
        An iris cube of the data to regrid, or multiple cubes to regrid in a
        cubelist. A minimum requirement is that the cube(s) need to be 2D with a
        latitude, longitude coordinates.
    xspacing: float
        Spacing of points in longitude direction (could be degrees, meters etc.)
    yspacing: float
        Spacing of points in latitude direction (could be degrees, meters etc.)
    method: str
        Method used to regrid onto, etc. Linear will use iris.analysis.Linear()

    Returns
    -------
    iris.cube | iris.cube.CubeList
        An iris cube of the data that has been regridded, or a cubelist of the
        cubes that have been regridded in the same order they were passed in
        toregrid.

    Raises
    ------
    ValueError
        If a unique x/y coordinate cannot be found
    NotImplementedError
        If the cubes grid, or the method for regridding, is not yet supported.

    Notes
    -----
    Currently rectlinear grids (uniform) are supported.

    """
    # To store regridded cubes.
    regridded_cubes = iris.cube.CubeList()

    # Iterate over all cubes and regrid.
    for cube in iter_maybe(toregrid):
        # Get x,y coord names
        y_coord, x_coord = get_cube_yxcoordname(cube)

        # List of supported grids - check if it is compatible
        supported_grids = (iris.coord_systems.GeogCS, iris.coord_systems.RotatedGeogCS)
        if not isinstance(cube.coord(x_coord).coord_system, supported_grids):
            raise NotImplementedError(
                f"Does not currently support {cube.coord(x_coord).coord_system} regrid method"
            )
        if not isinstance(cube.coord(y_coord).coord_system, supported_grids):
            raise NotImplementedError(
                f"Does not currently support {cube.coord(y_coord).coord_system} regrid method"
            )

        # Get axis
        lat, lon = cube.coord(y_coord), cube.coord(x_coord)

        # Get bounds
        lat_min, lon_min = lat.points.min(), lon.points.min()
        lat_max, lon_max = lat.points.max(), lon.points.max()

        # Generate new mesh
        latout = np.arange(lat_min, lat_max, yspacing)
        lonout = np.arange(lon_min, lon_max, xspacing)

        regrid_method = getattr(iris.analysis, method, None)
        if callable(regrid_method):
            regridded_cubes.append(
                cube.interpolate(
                    [(y_coord, latout), (x_coord, lonout)], regrid_method()
                )
            )
        else:
            raise NotImplementedError(
                f"Does not currently support {method} regrid method"
            )

    # Preserve returning a cube if only a cube has been supplied to regrid.
    if len(regridded_cubes) == 1:
        return regridded_cubes[0]
    else:
        return regridded_cubes


def regrid_to_single_point(
    cube: iris.cube.Cube,
    lat_pt: float,
    lon_pt: float,
    latlon_in_type: str = "rotated",
    method: str = "Nearest",
    boundary_margin: int = 8,
    **kwargs,
) -> iris.cube.Cube:
    """Select data at a single point by longitude and latitude.

    Selection of model grid point is performed by a regrid function, either
    selecting the nearest gridpoint to the selected longitude and latitude
    values or using linear interpolation across the surrounding points.

    Parameters
    ----------
    cube: Cube
        An iris cube of the data to regrid. As a minimum, it needs to be 2D with
        latitude, longitude coordinates.
    lon_pt: float
        Selected value of longitude: this should be in the range -180 degrees to
        180 degrees.
    lat_pt: float
        Selected value of latitude: this should be in the range -90 degrees to
        90 degrees.
    method: str
        Method used to determine the values at the selected longitude and
        latitude. The recommended approach is to use iris.analysis.Nearest(),
        which selects the nearest gridpoint. An alternative is
        iris.analysis.Linear(), which obtains the values at the selected
        longitude and latitude by linear interpolation.
    boundary_margin: int, optional
        Number of grid points from the domain boundary considered "unreliable".
        Defaults to 8.

    Returns
    -------
    cube_rgd: Cube
        An iris cube of the data at the specified point (this may have time
        and/or height dimensions).

    Raises
    ------
    ValueError
        If a unique x/y coordinate cannot be found; also if, for selecting a
        single gridpoint, the chosen longitude and latitude point is outside the
        domain.
    NotImplementedError
        If the cubes grid, or the method for regridding, is not yet supported.

    Notes
    -----
    The acceptable coordinate names for X and Y coordinates are currently
    described in X_COORD_NAMES and Y_COORD_NAMES. These cover commonly used
    coordinate types, though a user can append new ones. Currently rectilinear
    grids (uniform) are supported. Warnings are raised if the selected gridpoint
    is within boundary_margin grid lengths of the domain boundary as data here
    is potentially unreliable.

    """
    # Get x and y coordinate names.
    y_coord, x_coord = get_cube_yxcoordname(cube)

    # List of supported grids - check if it is compatible
    # NOTE: The "RotatedGeogCS" option below seems to be required for rotated grids --
    #  this may need to be added in other places in these Operators.
    supported_grids = (iris.coord_systems.GeogCS, iris.coord_systems.RotatedGeogCS)
    if not isinstance(cube.coord(x_coord).coord_system, supported_grids):
        raise NotImplementedError(
            f"Does not currently support {cube.coord(x_coord).coord_system} regrid method"
        )
    if not isinstance(cube.coord(y_coord).coord_system, supported_grids):
        raise NotImplementedError(
            f"Does not currently support {cube.coord(y_coord).coord_system} regrid method"
        )

    # Transform input coordinates onto rotated grid if requested
    if latlon_in_type == "realworld":
        lon_tr, lat_tr = transform_lat_long_points(lon_pt, lat_pt, cube)
    elif latlon_in_type == "rotated":
        lon_tr, lat_tr = lon_pt, lat_pt

    # Get axis
    lat, lon = cube.coord(y_coord), cube.coord(x_coord)

    # Get bounds
    lat_min, lon_min = lat.points.min(), lon.points.min()
    lat_max, lon_max = lat.points.max(), lon.points.max()

    # Get boundaries of frame to avoid selecting gridpoint close to domain edge
    lat_min_bound, lon_min_bound = (
        lat.points[boundary_margin - 1],
        lon.points[boundary_margin - 1],
    )
    lat_max_bound, lon_max_bound = (
        lat.points[-boundary_margin],
        lon.points[-boundary_margin],
    )

    # Check to see if selected point is outside the domain
    if (lat_tr < lat_min) or (lat_tr > lat_max):
        raise ValueError("Selected point is outside the domain.")
    else:
        if (lon_tr < lon_min) or (lon_tr > lon_max):
            if (lon_tr + 360.0 >= lon_min) and (lon_tr + 360.0 <= lon_max):
                lon_tr += 360.0
            elif (lon_tr - 360.0 >= lon_min) and (lon_tr - 360.0 <= lon_max):
                lon_tr -= 360.0
            else:
                raise ValueError("Selected point is outside the domain.")

    # Check to see if selected point is near the domain boundaries
    if (
        (lat_tr < lat_min_bound)
        or (lat_tr > lat_max_bound)
        or (lon_tr < lon_min_bound)
        or (lon_tr > lon_max_bound)
    ):
        warnings.warn(
            f"Selected point is within {boundary_margin} gridlengths of the domain edge, data may be unreliable.",
            category=BoundaryWarning,
            stacklevel=2,
        )

    regrid_method = getattr(iris.analysis, method, None)
    if not callable(regrid_method):
        raise NotImplementedError(f"Does not currently support {method} regrid method")
    cube_rgd = cube.interpolate(((lat, lat_tr), (lon, lon_tr)), regrid_method())
    return cube_rgd


def transform_lat_long_points(lon, lat, cube):
    """Transform a selected point in longitude and latitude.

    Transform the coordinates of a point from the real world
    grid to the corresponding point on the rotated grid of a cube.

    Parameters
    ----------
    cube: Cube
        An iris cube of data defining the rotated grid to be used in
        the longitude-latitude transformation.
    lon: float
        Selected value of longitude: this should be in the range -180 degrees to
        180 degrees.
    lat: float
        Selected value of latitude: this should be in the range -90 degrees to
        90 degrees.

    Returns
    -------
    lon_rot, lat_rot: float
        Coordinates of the selected point on the rotated grid specified within
        the selected cube.

    """
    import cartopy.crs as ccrs

    rot_pole = cube.coord_system().as_cartopy_crs()
    true_grid = ccrs.Geodetic()
    rot_coords = rot_pole.transform_point(lon, lat, true_grid)
    lon_rot = rot_coords[0]
    lat_rot = rot_coords[1]

    return lon_rot, lat_rot



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators to aggregate across either 1 or 2 dimensions."""

import iris
import iris.analysis
import iris.coord_categorisation
import iris.cube
import isodate

from CSET._common import iter_maybe
from CSET.operators._utils import is_time_aggregatable


def time_aggregate(
    cube: iris.cube.Cube,
    method: str,
    interval_iso: str,
    **kwargs,
) -> iris.cube.Cube:
    """Aggregate cube by its time coordinate.

    Aggregates similar (stash) fields in a cube for the specified coordinate and
    using the method supplied. The aggregated cube will keep the coordinate and
    add a further coordinate with the aggregated end time points.

    Examples are: 1. Generating hourly or 6-hourly precipitation accumulations
    given an interval for the new time coordinate.

    We use the isodate class to convert ISO 8601 durations into time intervals
    for creating a new time coordinate for aggregation.

    We use the lambda function to pass coord and interval into the callable
    category function in add_categorised to allow users to define their own
    sub-daily intervals for the new time coordinate.

    Arguments
    ---------
    cube: iris.cube.Cube
        Cube to aggregate and iterate over one dimension
    coordinate: str
        Coordinate to aggregate over i.e. 'time', 'longitude',
        'latitude','model_level_number'.
    method: str
        Type of aggregate i.e. method: 'SUM', getattr creates
        iris.analysis.SUM, etc.
    interval_iso: isodate timedelta ISO 8601 object i.e PT6H (6 hours), PT30M (30 mins)
        Interval to aggregate over.

    Returns
    -------
    cube: iris.cube.Cube
        Single variable but several methods of aggregation

    Raises
    ------
    ValueError
        If the constraint doesn't produce a single cube containing a field.
    """
    # Duration of ISO timedelta.
    timedelta = isodate.parse_duration(interval_iso)

    # Convert interval format to whole hours.
    interval = int(timedelta.total_seconds() / 3600)

    # Add time categorisation overwriting hourly increment via lambda coord.
    # https://scitools-iris.readthedocs.io/en/latest/_modules/iris/coord_categorisation.html
    iris.coord_categorisation.add_categorised_coord(
        cube, "interval", "time", lambda coord, cell: cell // interval * interval
    )

    # Aggregate cube using supplied method.
    aggregated_cube = cube.aggregated_by("interval", getattr(iris.analysis, method))
    aggregated_cube.remove_coord("interval")
    return aggregated_cube


def ensure_aggregatable_across_cases(
    cube: iris.cube.Cube | iris.cube.CubeList,
) -> iris.cube.Cube:
    """Ensure a Cube or CubeList can be aggregated across multiple cases.

    Arguments
    ---------
    cube: iris.cube.Cube | iris.cube.CubeList
        If a Cube is provided it is checked to determine if it has the
        the necessary dimensional coordinates to be aggregatable.
        These necessary coordinates are 'forecast_period' and
        'forecast_reference_time'.If a CubeList is provided a Cube is created
        by slicing over all time coordinates and the resulting list is merged
        to create an aggregatable cube.

    Returns
    -------
    cube: iris.cube.Cube
        A time aggregatable cube with dimension coordinates including
        'forecast_period' and 'forecast_reference_time'.

    Raises
    ------
    ValueError
        If a Cube is provided and it is not aggregatable a ValueError is
        raised. The user should then provide a CubeList to be turned into an
        aggregatable cube to allow aggregation across multiple cases to occur.

    Notes
    -----
    This is a simple operator designed to ensure that a Cube is aggregatable
    across cases. If a CubeList is presented it will create an aggregatable Cube
    from that list. Its functionality is for case study (or trial) aggregation
    to ensure that the full dataset can be loaded as a single cube. This
    functionality is particularly useful for percentiles, Q-Q plots, and
    histograms.
    """
    # Check to see if a cube is input and if that cube is iterable.
    if isinstance(cube, iris.cube.Cube):
        if is_time_aggregatable(cube):
            return cube
        else:
            raise ValueError(
                "Single Cube should have 'forecast_period' and"
                "'forecast_reference_time' dimensional coordinates. "
                "To make a time aggregatable Cube input a CubeList."
            )
    # Create an aggregatable cube from the provided CubeList.
    else:
        new_cube_list = iris.cube.CubeList()
        for sub_cube in cube:
            for cube_slice in sub_cube.slices_over(
                ["forecast_period", "forecast_reference_time"]
            ):
                new_cube_list.append(cube_slice)
        new_merged_cube = new_cube_list.merge_cube()
        return new_merged_cube


def add_hour_coordinate(
    cubes: iris.cube.Cube | iris.cube.CubeList,
) -> iris.cube.Cube | iris.cube.CubeList:
    """Add a category coordinate of hour of day to a Cube or CubeList.

    Arguments
    ---------
    cubes: iris.cube.Cube | iris.cube.CubeList
        Cube of any variable that has a time coordinate.

    Returns
    -------
    cube: iris.cube.Cube
        A Cube with an additional auxiliary coordinate of hour.

    Notes
    -----
    This is a simple operator designed to be used prior to case aggregation for
    histograms, Q-Q plots, and percentiles when aggregated by hour of day.
    """
    new_cubelist = iris.cube.CubeList()
    for cube in iter_maybe(cubes):
        # Add a category coordinate of hour into each cube.
        iris.coord_categorisation.add_hour(cube, "time", name="hour")
        cube.coord("hour").units = "hours"
        new_cubelist.append(cube)

    if len(new_cubelist) == 1:
        return new_cubelist[0]
    else:
        return new_cubelist



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common functionality used across CSET."""

import ast
import io
import json
import logging
import re
from collections.abc import Iterable
from pathlib import Path

import ruamel.yaml


class ArgumentError(ValueError):
    """Provided arguments are not understood."""


def parse_recipe(recipe_yaml: Path | str, variables: dict = None) -> dict:
    """Parse a recipe into a python dictionary.

    Parameters
    ----------
    recipe_yaml: Path | str
        Path to recipe file, or the recipe YAML directly.
    variables: dict
        Dictionary of recipe variables. If None templating is not attempted.

    Returns
    -------
    recipe: dict
        The recipe as a python dictionary.

    Raises
    ------
    ValueError
        If the recipe is invalid. E.g. invalid YAML, missing any steps, etc.
    TypeError
        If recipe_yaml isn't a Path or string.
    KeyError
        If needed recipe variables are not supplied.

    Examples
    --------
    >>> CSET._common.parse_recipe(Path("myrecipe.yaml"))
    {'steps': [{'operator': 'misc.noop'}]}
    """
    # Ensure recipe_yaml is something the YAML parser can read.
    if isinstance(recipe_yaml, str):
        recipe_yaml = io.StringIO(recipe_yaml)
    elif not isinstance(recipe_yaml, Path):
        raise TypeError("recipe_yaml must be a str or Path.")

    # Parse the recipe YAML.
    with ruamel.yaml.YAML(typ="safe", pure=True) as yaml:
        try:
            recipe = yaml.load(recipe_yaml)
        except ruamel.yaml.parser.ParserError as err:
            raise ValueError("ParserError: Invalid YAML") from err

    logging.debug("Recipe before templating:\n%s", recipe)
    check_recipe_has_steps(recipe)

    if variables is not None:
        logging.debug("Recipe variables: %s", variables)
        recipe = template_variables(recipe, variables)

    logging.debug("Recipe after templating:\n%s", recipe)
    return recipe


def check_recipe_has_steps(recipe: dict):
    """Check a recipe has the minimum required steps.

    Checking that the recipe actually has some steps, and providing helpful
    error messages otherwise. We must have at least a steps step, as that
    reads the raw data.

    Parameters
    ----------
    recipe: dict
        The recipe as a python dictionary.

    Raises
    ------
    ValueError
        If the recipe is invalid. E.g. invalid YAML, missing any steps, etc.
    TypeError
        If recipe isn't a dict.
    KeyError
        If needed recipe variables are not supplied.
    """
    if not isinstance(recipe, dict):
        raise TypeError("Recipe must contain a mapping.")
    if "steps" not in recipe:
        raise ValueError("Recipe must contain a 'steps' key.")
    try:
        if len(recipe["steps"]) < 1:
            raise ValueError("Recipe must have at least 1 step.")
    except TypeError as err:
        raise ValueError("'steps' key must contain a sequence of steps.") from err


def slugify(s: str) -> str:
    """Turn a string into a version that can be used everywhere.

    The resultant string will only consist of a-z, 0-9, dots, dashes, and
    underscores.
    """
    return re.sub(r"[^a-z0-9\._-]+", "_", s.casefold()).strip("_")


def get_recipe_metadata() -> dict:
    """Get the metadata of the running recipe."""
    try:
        with open("meta.json", "rt", encoding="UTF-8") as fp:
            return json.load(fp)
    except FileNotFoundError:
        meta = {}
        with open("meta.json", "wt", encoding="UTF-8") as fp:
            json.dump(meta, fp, indent=2)
        return {}


def parse_variable_options(arguments: list[str]) -> dict:
    """Parse a list of arguments into a dictionary of variables.

    The variable name arguments start with two hyphen-minus (`--`), consisting
    of only capital letters (`A`-`Z`) and underscores (`_`). While the variable
    name is restricted, the value of the variable can be any string.

    Parameters
    ----------
    arguments: list[str]
        List of arguments, e.g: `["--LEVEL", "2", "--STASH=m01s01i001"]`

    Returns
    -------
    recipe_variables: dict
        Dictionary keyed with the variable names.

    Raises
    ------
    ValueError
        If any arguments cannot be parsed.
    """
    recipe_variables = {}
    i = 0
    while i < len(arguments):
        if re.fullmatch(r"--[A-Z_]+=.*", arguments[i]):
            key, value = arguments[i].split("=", 1)
        elif re.fullmatch(r"--[A-Z_]+", arguments[i]):
            try:
                key = arguments[i].strip("-")
                value = arguments[i + 1]
            except IndexError as err:
                raise ArgumentError(f"No value for variable {arguments[i]}") from err
            i += 1
        else:
            raise ArgumentError(f"Unknown argument: {arguments[i]}")
        try:
            # Remove quotes from arguments, in case left in CSET_ADDOPTS.
            if re.fullmatch(r"""["'].+["']""", value):
                value = value[1:-1]
            recipe_variables[key.strip("-")] = ast.literal_eval(value)
        # Capture the many possible exceptions from ast.literal_eval
        except (ValueError, TypeError, SyntaxError, MemoryError, RecursionError):
            recipe_variables[key.strip("-")] = value
        i += 1
    return recipe_variables


def template_variables(recipe: dict | list, variables: dict) -> dict:
    """Insert variables into recipe.

    Parameters
    ----------
    recipe: dict | list
        The recipe as a python dictionary. It is updated in-place.
    variables: dict
        Dictionary of variables for the recipe.

    Returns
    -------
    recipe: dict
        Filled recipe as a python dictionary.

    Raises
    ------
    KeyError
        If needed recipe variables are not supplied.
    """
    if isinstance(recipe, dict):
        index = recipe.keys()
    elif isinstance(recipe, list):
        # We have to handle lists for when we have one inside a recipe.
        index = range(len(recipe))
    else:
        raise TypeError("recipe must be a dict or list.", recipe)

    for i in index:
        if isinstance(recipe[i], (dict, list)):
            recipe[i] = template_variables(recipe[i], variables)
        elif isinstance(recipe[i], str):
            recipe[i] = replace_template_variable(recipe[i], variables)
    return recipe


def replace_template_variable(s: str, variables):
    """Fill all variable placeholders in the string."""
    for var_name, var_value in variables.items():
        placeholder = f"${var_name}"
        # If the value is just the placeholder we directly overwrite it
        # to keep the value type.
        if s == placeholder:
            s = var_value
            break
        else:
            s = s.replace(placeholder, str(var_value))
    if isinstance(s, str) and re.match(r"^.*\$[A-Z_].*", s):
        raise KeyError("Variable without a value.", s)
    return s


################################################################################
# Templating code taken from the simple_template package under the 0BSD licence.
# Original at https://github.com/Fraetor/simple_template
################################################################################


class TemplateError(KeyError):
    """Rendering a template failed due a placeholder without a value."""


def render(template: str, /, **variables) -> str:
    """Render the template with the provided variables.

    The template should contain placeholders that will be replaced. These
    placeholders consist of the placeholder name within double curly braces. The
    name of the placeholder should be a valid python identifier. Whitespace
    between the braces and the name is ignored. E.g.: `{{ placeholder_name }}`

    An exception will be raised if there are placeholders without corresponding
    values. It is acceptable to provide unused values; they will be ignored.

    Parameters
    ----------
    template: str
        Template to fill with variables.

    **variables: Any
        Keyword arguments for the placeholder values. The argument name should
        be the same as the placeholder's name. You can unpack a dictionary of
        value with `render(template, **my_dict)`.

    Returns
    -------
    rendered_template: str
        Filled template.

    Raises
    ------
    TemplateError
        Value not given for a placeholder in the template.
    TypeError
        If the template is not a string, or a variable cannot be casted to a
        string.

    Examples
    --------
    >>> template = "<p>Hello {{myplaceholder}}!</p>"
    >>> simple_template.render(template, myplaceholder="World")
    "<p>Hello World!</p>"
    """

    def isidentifier(s: str):
        return s.isidentifier()

    def extract_placeholders():
        matches = re.finditer(r"{{\s*([^}]+)\s*}}", template)
        unique_names = {match.group(1) for match in matches}
        return filter(isidentifier, unique_names)

    def substitute_placeholder(name):
        try:
            value = str(variables[name])
        except KeyError as err:
            raise TemplateError("Placeholder missing value", name) from err
        pattern = r"{{\s*%s\s*}}" % re.escape(name)
        return re.sub(pattern, value, template)

    for name in extract_placeholders():
        template = substitute_placeholder(name)
    return template


def render_file(template_path: str, /, **variables) -> str:
    """Render a template directly from a file.

    Otherwise the same as `simple_template.render()`.

    Examples
    --------
    >>> simple_template.render_file("/path/to/template.html", myplaceholder="World")
    "<p>Hello World!</p>"
    """
    with open(template_path, "rt", encoding="UTF-8") as fp:
        template = fp.read()
    return render(template, **variables)


def iter_maybe(thing) -> Iterable:
    """Ensure thing is Iterable. Strings count as atoms."""
    if isinstance(thing, Iterable) and not isinstance(thing, str):
        return thing
    return (thing,)


def combine_dicts(d1: dict, d2: dict) -> dict:
    """Recursively combines two dictionaries.

    Duplicate atoms favour the second dictionary.
    """
    # Update existing keys.
    for key in d1.keys() & d2.keys():
        if isinstance(d1[key], dict):
            d1[key] = combine_dicts(d1[key], d2[key])
        else:
            d1[key] = d2[key]
    # Add any new keys.
    for key in d2.keys() - d1.keys():
        d1[key] = d2[key]
    return d1



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for common functionality across CSET."""

from collections.abc import Iterable
from pathlib import Path

import pytest

import CSET._common as common


def test_parse_recipe_string():
    """Loading and parsing of a YAML recipe from a string."""
    valid_recipe = """\
    steps:
        operator: misc.noop
        arg1: Hello
    """
    parsed = common.parse_recipe(valid_recipe)
    assert parsed == {"steps": {"operator": "misc.noop", "arg1": "Hello"}}


def test_parse_recipe_path():
    """Loading and parsing of a YAML recipe from a Path."""
    parsed = common.parse_recipe(Path("tests/test_data/noop_recipe.yaml"))
    expected = {
        "title": "Noop",
        "description": "A recipe that does nothing. Only used for testing.",
        "steps": [
            {
                "operator": "misc.noop",
                "test_argument": "Banana",
                "dict_argument": {"key": "value"},
                "substep": {"operator": "constraints.combine_constraints"},
            }
        ],
    }
    assert parsed == expected


def test_parse_recipe_insert_variables():
    """Test insertion of variables into a recipe."""
    recipe_yaml = '{"steps": [{"operator": "misc.noop", "argument": "$VALUE"}]}'
    variables = {"VALUE": 42}
    parsed = common.parse_recipe(recipe_yaml, variables)
    expected = {"steps": [{"operator": "misc.noop", "argument": 42}]}
    assert parsed == expected


def test_parse_recipe_exception_missing():
    """Exception for non-existent file."""
    with pytest.raises(FileNotFoundError):
        common.parse_recipe(Path("/non-existent/path"))


def test_parse_recipe_exception_type():
    """Exception for incorrect type."""
    with pytest.raises(TypeError, match="recipe_yaml must be a str or Path."):
        common.parse_recipe(True)


def test_parse_recipe_exception_invalid_yaml():
    """Exception for invalid YAML."""
    with pytest.raises(ValueError, match="ParserError: Invalid YAML"):
        common.parse_recipe('"Inside quotes" outside of quotes')


def test_parse_recipe_exception_invalid_recipe():
    """Exception for valid YAML but invalid recipe."""
    with pytest.raises(ValueError, match="Recipe must contain a 'steps' key."):
        common.parse_recipe("a: 1")


def test_parse_recipe_exception_blank():
    """Exception for blank recipe."""
    with pytest.raises(TypeError, match="Recipe must contain a mapping."):
        common.parse_recipe("")


def test_parse_recipe_exception_no_steps():
    """Exception for recipe without any steps steps."""
    with pytest.raises(ValueError, match="Recipe must have at least 1 step"):
        common.parse_recipe("steps: []")


def test_parse_recipe_exception_steps_not_sequence():
    """Exception for recipe with steps containing an atom."""
    with pytest.raises(
        ValueError, match="'steps' key must contain a sequence of steps."
    ):
        common.parse_recipe("steps: 7")


def test_parse_recipe_exception_non_dict():
    """Exception for recipe that parses to a non-dict."""
    with pytest.raises(TypeError, match="Recipe must contain a mapping."):
        common.parse_recipe("[]")


def test_slugify():
    """Slugify removes special characters."""
    assert common.slugify("Test") == "test"
    assert (
        common.slugify("Mean Surface Air Temperature Spatial Plot")
        == "mean_surface_air_temperature_spatial_plot"
    )
    assert common.slugify("file-name.yaml") == "file-name.yaml"
    assert common.slugify("First Line\nSecond Line") == "first_line_second_line"
    assert common.slugify("greekαβγδchars") == "greek_chars"
    assert common.slugify("  ABC ") == "abc"
    # Multi-byte unicode characters are removed.
    assert common.slugify("あいうえお") == ""


def test_parse_variable_options():
    """Variable arguments are parsed correctly."""
    args = ("--STASH=m01s01i001", "--COUNT", "3", "--CELL_METHODS=[]")
    expected = {"STASH": "m01s01i001", "COUNT": 3, "CELL_METHODS": []}
    actual = common.parse_variable_options(args)
    assert actual == expected
    # Not valid variable name.
    with pytest.raises(ValueError):
        common.parse_variable_options(("--lowercase=False",))
    # Missing variable value.
    with pytest.raises(ValueError):
        common.parse_variable_options(("--VARIABLE",))


def test_parse_variable_options_quoted():
    """Quoted arguments get unquoted before interpretation."""
    args = ["--A='None'", "--B", '"None"']
    expected = {"A": None, "B": None}
    actual = common.parse_variable_options(args)
    assert actual == expected


def test_template_variables():
    """Multiple variables are correctly templated into recipe."""
    recipe = {
        "steps": [{"operator": "misc.noop", "v1": "$VAR_A", "v2": "$VAR_B", "v3": 0}]
    }
    variables = {"VAR_A": 42, "VAR_B": 3.14}
    expected = {"steps": [{"operator": "misc.noop", "v1": 42, "v2": 3.14, "v3": 0}]}
    actual = common.template_variables(recipe, variables)
    assert actual == expected
    assert recipe == expected


def test_template_variables_wrong_recipe_type():
    """Give wrong type for recipe."""
    with pytest.raises(TypeError):
        common.template_variables(1, {})


def test_replace_template_variable():
    """Placeholders are correctly substituted."""
    # Test direct substitution.
    vars = {"VAR": 1}
    expected = 1
    actual = common.replace_template_variable("$VAR", vars)
    assert actual == expected

    # Insertion into a larger string.
    expected = "The number 1"
    actual = common.replace_template_variable("The number $VAR", vars)
    assert actual == expected

    # Error when variable not provided.
    with pytest.raises(KeyError):
        common.replace_template_variable("$VAR", {})


def test_get_recipe_meta(tmp_working_dir):
    """Reading metadata from disk."""
    # Default for missing file.
    meta_file = Path("meta.json")
    assert not meta_file.exists()
    assert common.get_recipe_metadata() == {}
    assert meta_file.exists()
    # Reads existing file.
    meta_file.write_text('{"title": "Example Title"}', encoding="UTF-8")
    assert common.get_recipe_metadata()["title"] == "Example Title"


def test_simple_placeholder():
    """Simple case of a single placeholder being templated."""
    template = "<p>{{greeting}} World!</p>"
    actual = common.render(template, greeting="Hello")
    expected = "<p>Hello World!</p>"
    assert actual == expected


def test_multiple_placeholders():
    """Templating of multiple different placeholders."""
    template = "<p>{{greeting}} {{who}}!</p>"
    actual = common.render(template, greeting="Hello", who="World")
    expected = "<p>Hello World!</p>"
    assert actual == expected


def test_repeated_placeholders():
    """Templating of repeated placeholders."""
    template = "<p>{{who}} says {{greeting}} {{ who }}!</p>"
    actual = common.render(template, greeting="Hello", who="World")
    expected = "<p>World says Hello World!</p>"
    assert actual == expected


def test_template_with_no_placeholders():
    """Trivial case with no placeholders."""
    template = "<p>Hello World!</p>"
    assert common.render(template) == template


def test_extra_arguments():
    """Give unneeded keyword arguments to ensure they are ignored."""
    template = "<p>{{greeting}} World!</p>"
    actual = common.render(template, greeting="Hello", extra="Ignored")
    expected = "<p>Hello World!</p>"
    assert actual == expected


def test_missing_argument_exception():
    """Exception raised if template is missing a value."""
    template = "<p>{{greeting}} World!</p>"
    with pytest.raises(common.TemplateError):
        common.render(template)


def test_improper_template_type_exception():
    """Exception raised if template isn't a string."""
    template = 3.14159265
    with pytest.raises(TypeError):
        common.render(template)


def test_render_file():
    """Render a template in a file."""
    actual = common.render_file("tests/test_data/template_file.html", greeting="Hello")
    expected = "<p>Hello World!</p>\n"
    assert actual == expected


def test_iter_maybe_iterable():
    """Pass an iterable unchanged though iter_maybe."""
    actual = [1, 2, 3]
    expected = common.iter_maybe(actual)
    # The same object is returned.
    assert expected is actual


def test_iter_maybe_atom():
    """Convert an atom into an iterable."""
    atom = 7
    created_iterable = common.iter_maybe(atom)
    assert isinstance(created_iterable, Iterable)
    for value in created_iterable:
        # The same object is inside the iterable.
        assert value is atom


def test_iter_maybe_string():
    """String is wrapped inside of an iterable instead being a char iterator."""
    atom = "A string"
    created_iterable = common.iter_maybe(atom)
    assert isinstance(created_iterable, Iterable)
    for value in created_iterable:
        # The same object is inside the iterable.
        assert value is atom


def test_combine_dicts():
    """Test combine_dicts function."""
    d1 = {"a": 1, "b": 2, "c": {"d": 3, "e": 4}}
    d2 = {"b": 3, "c": {"d": 5, "f": 6}}
    expected = {"a": 1, "b": 3, "c": {"d": 5, "e": 4, "f": 6}}
    assert common.combine_dicts(d1, d2) == expected



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Graph tests."""

import os
import stat
from pathlib import Path

import pytest

from CSET import graph


def test_save_graph():
    """Generate a graph from a recipe file without the output specified."""
    graph.save_graph(Path("tests/test_data/plot_instant_air_temp.yaml"))


def test_save_graph_detailed(tmp_path: Path):
    """Generate a graph from a recipe file with the output specified."""
    graph.save_graph(
        Path("tests/test_data/plot_instant_air_temp.yaml"),
        tmp_path / "graph.svg",
        detailed=True,
    )


def test_save_graph_no_operators_exception():
    """Exception raised from recipe with no operators in its steps."""
    with pytest.raises(ValueError):
        # Inline YAML form used.
        graph.save_graph('{"steps": [{"argument": "no_operators"}]}')


def test_save_graph_no_steps_exception():
    """Exception raised from recipe with no steps."""
    with pytest.raises(ValueError):
        graph.save_graph("title: Recipe with no steps")


def test_save_graph_auto_open_xdg_open(tmp_path: Path, monkeypatch):
    """Test the auto-opening of the graph with (a mocked) xdg-open."""
    xdg_open = tmp_path / "xdg-open"
    with open(xdg_open, "wt", encoding="UTF-8") as fp:
        fp.write("#!/bin/bash\ntrue")
    xdg_open.chmod(stat.S_IRUSR | stat.S_IXUSR)
    monkeypatch.setenv("PATH", str(tmp_path), prepend=os.pathsep)
    graph.save_graph(Path("tests/test_data/plot_instant_air_temp.yaml"), auto_open=True)


def test_save_graph_auto_open_no_xdg_open(tmp_path: Path, monkeypatch):
    """Test the auto-opening of the graph errors without xdg-open.

    Strictly speaking this tests xdg-open failing, but it is equivalent to it
    missing.
    """
    xdg_open = tmp_path / "xdg-open"
    with open(xdg_open, "wt", encoding="UTF-8") as fp:
        fp.write("#!/bin/bash\nfalse")
    xdg_open.chmod(stat.S_IRUSR | stat.S_IXUSR)
    monkeypatch.setenv("PATH", str(tmp_path), prepend=os.pathsep)
    graph.save_graph(Path("tests/test_data/plot_instant_air_temp.yaml"), auto_open=True)



# © Crown copyright, Met Office (2022-2025) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module containing different diagnostics for ensembles.

The diagnostics here are applicable to ensembles and apply generally to
ensembles. They are not just limited to considering ensemble spread.
"""

import iris
import iris.cube
import numpy as np


def DKE(u: iris.cube.Cube, v: iris.cube.Cube) -> iris.cube.Cube:
    r"""Calculate the Difference Kinetic Energy (DKE).

    Parameters
    ----------
    u: iris.cube.Cube
        Iris cube of the u component of the wind field for all members. Must
        include a realization coordinate; realization must be the first
        coordinate.
    v: iris.cube.Cube
        Iris cube of the v component of the wind field same format as u.

    Returns
    -------
    DKE: iris.cube.Cube
        An iris cube of the DKE for each of the control - member comparisons.

    Notes
    -----
    The Difference Kinetic Energy, or DKE was first used in an ensemble sense
    by [Zhangetal2002]_. It is calculated as

    .. math:: \frac{1}{2}u'u' + \frac{1}{2}v'v'

    where

    .. math:: x' = x_{control} - x_{perturbed}

    for each member of the ensemble.

    The DKE is used to show links between the perturbation growth (growth of
    ensemble spread) or errors with dynamical features. It can be particularly
    useful in understanding differences in physical mechanisms between ensemble
    members. The larger the DKE the greater the difference between the two
    members being considered.

    The DKE can be viewed as a domain average, horizontal integration (to
    produce profiles), or vertical integration/subsampling (to provide spatial maps).
    Furthermore, weightings based on volume or mass can be applied to these
    integrations. However, initially the DKE per unit mass is implemented here.

    The DKE is often considered in the form of power spectra to identify the
    presence of upscale error growth or the dominant scales of error growth. It
    is often plotted on a logarithmic scale.

    References
    ----------
    .. [Zhangetal2002] Zhang, F., Snyder, C., and Rotunno, R., (2002)
       "Mesoscale Predictability of the 'Surprise' Snowstorm of 24-25 January
       2000." Monthly Weather Review, vol. 130, 1617-1632,
       doi: 10.1175/1520-0493(2002)130<1617:MPOTSS>2.0.CO;2

    Examples
    --------
    >>> DKE = ensembles.DKE(u, v)
    """
    for cube in [u, v]:
        # Use dim_map to store the dimensions and check the realization
        # coordinate is the first coordinate.
        dim_map = {}
        for coord in cube.dim_coords:
            dim_index = cube.coord_dims(coord.name())[0]
            dim_map[coord.name()] = dim_index
        if "realization" not in dim_map:
            raise ValueError("Cube should have a realization coordinate.")
        if not dim_map["realization"] == 0:
            raise ValueError("Realization should be the first coordinate in cube.")

    # Check cubes are the same shape.
    if u.shape != v.shape:
        raise ValueError("Cubes should be the same shape.")

    # Check coordinates are identical.
    for coord_u, coord_v in zip(u.dim_coords, v.dim_coords, strict=True):
        if not u.coord(coord_u.name()) == v.coord(coord_v.name()):
            raise ValueError(
                f"u and v should have matching coordinates for {coord_u.name()}."
            )
    # Define control member and perturbed members.
    u_ctrl = u[np.where(u.coord("realization").points[:] == 0)[0][0], :]
    u_mem = u[np.where(u.coord("realization").points[:] != 0)[0][:], :]
    v_ctrl = v[np.where(v.coord("realization").points[:] == 0)[0][0], :]
    v_mem = v[np.where(v.coord("realization").points[:] != 0)[0][:], :]

    # Calculate the DKE
    DKE = u_mem.copy()
    DKE.data[:] = 0.0
    DKE.data = (
        0.5 * (u_ctrl.data - u_mem.data) ** 2 + 0.5 * (v_ctrl.data - v_mem.data) ** 2
    )
    DKE.rename("Difference Kinetic Energy")
    return DKE



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test constraint operators."""

from datetime import datetime

import pytest

from CSET.operators import constraints


def test_generate_stash_constraint():
    """Generate iris cube constraint for UM STASH code."""
    stash_constraint = constraints.generate_stash_constraint("m01s03i236")
    expected_stash_constraint = "AttributeConstraint({'STASH': 'm01s03i236'})"
    assert repr(stash_constraint) == expected_stash_constraint


def test_generate_var_constraint():
    """Generate iris cube constraint for str variable name."""
    var_constraint = constraints.generate_var_constraint("test")
    expected_var_constraint = "Constraint(name='test')"
    assert repr(var_constraint) == expected_var_constraint


def test_generate_var_constraint_stash():
    """Generate iris cube constraint for UM STASH code with var constraint."""
    var_constraint = constraints.generate_var_constraint("m01s03i236")
    expected_stash_constraint = "AttributeConstraint({'STASH': 'm01s03i236'})"
    assert repr(var_constraint) == expected_stash_constraint


def test_generate_cell_methods_constraint():
    """Generate iris cube constraint for cell methods."""
    cell_methods_constraint = constraints.generate_cell_methods_constraint(["mean"])
    expected_cell_methods_constraint = "Constraint(cube_func=<function generate_cell_methods_constraint.<locals>.check_cell_methods at"
    assert expected_cell_methods_constraint in repr(cell_methods_constraint)


def test_generate_cell_methods_constraint_no_aggregation():
    """Generate iris cube constraint for no aggregation cell methods."""
    cell_methods_constraint = constraints.generate_cell_methods_constraint([])
    expected_cell_methods_constraint = "Constraint(cube_func=<function generate_cell_methods_constraint.<locals>.check_no_aggregation at"
    assert expected_cell_methods_constraint in repr(cell_methods_constraint)


def test_generate_time_constraint():
    """Generate iris cube constraint for dates."""
    # Try with str dates
    time_constraint = constraints.generate_time_constraint(
        "2023-03-24T00:00", "2023-03-24T06:00"
    )
    expected_time_constraint = "Constraint(coord_values={'time': <function generate_time_constraint.<locals>.<lambda> at "
    assert expected_time_constraint in repr(time_constraint)
    # Try with datetime.datetime dates
    time_constraint = constraints.generate_time_constraint(
        datetime.fromisoformat("2023-03-24T00:00:00+00:00"),
        datetime.fromisoformat("2023-03-24T06:00:00+00:00"),
    )
    assert expected_time_constraint in repr(time_constraint)
    # Try with implicit end
    time_constraint = constraints.generate_time_constraint("2023-03-24T00:00:00+00:00")
    assert expected_time_constraint in repr(time_constraint)


def test_generate_level_constraint_single_level():
    """Generate constraint for a single level."""
    pressure_constraint = constraints.generate_level_constraint(
        coordinate="pressure", levels=1000
    )
    expected_pressure_constraint = "Constraint(coord_values={'pressure': [1000]})"
    assert expected_pressure_constraint in repr(pressure_constraint)


def test_generate_level_constraint_multi_level():
    """Generate constraint for multiple pressure levels."""
    pressure_constraint = constraints.generate_level_constraint(
        coordinate="pressure", levels=[200, 800]
    )
    expected_pressure_constraint = "Constraint(coord_values={'pressure': [200, 800]})"
    assert expected_pressure_constraint in repr(pressure_constraint)


def test_generate_level_constraint_all_level():
    """Generate constraint for all levels."""
    pressure_constraint = constraints.generate_level_constraint(
        coordinate="pressure", levels="*"
    )
    expected_pressure_constraint = "Constraint(coord_values={'pressure': <function generate_level_constraint.<locals>.<lambda> at"
    assert expected_pressure_constraint in repr(pressure_constraint)


def test_generate_level_constraint_no_pressure():
    """Generate constraint for not having pressure levels."""
    pressure_constraint = constraints.generate_level_constraint(
        coordinate="pressure", levels=[]
    )
    expected_pressure_constraint = (
        "Constraint(cube_func=<function generate_level_constraint.<locals>.no_levels at"
    )
    assert expected_pressure_constraint in repr(pressure_constraint)


def test_generate_area_constraint():
    """Generate area constraint with lat-lon limits."""
    area_constraint = constraints.generate_area_constraint(0.0, 0.0, 0.1, 0.1)
    actual = repr(area_constraint)
    assert "Constraint(coord_values={" in actual
    assert (
        "'grid_latitude': <function generate_area_constraint.<locals>.bound_lat at 0x"
        in actual
    )
    assert (
        "'grid_longitude': <function generate_area_constraint.<locals>.bound_lon at 0x"
        in actual
    )


def test_generate_area_constraint_no_limits():
    """Generate area constraint with no limits."""
    area_constraint = constraints.generate_area_constraint(None, None, None, None)
    expected_area_constraint = "Constraint()"
    assert expected_area_constraint in repr(area_constraint)


def test_generate_area_constraint_invalid_arguments():
    """Generate area constraint raises exception with invalid arguments."""
    # Non-numbers are rejected.
    with pytest.raises(TypeError):
        constraints.generate_area_constraint(1, 2, 3, "four")

    # Mixed numbers and Nones are rejected.
    with pytest.raises(TypeError):
        constraints.generate_area_constraint(None, None, None, 0)


def test_combine_constraints():
    """Combine constraint."""
    stash_constraint = constraints.generate_stash_constraint("m01s03i236")
    var_constraint = constraints.generate_var_constraint("test")
    combined_constraint = constraints.combine_constraints(
        stash_constraint,
        a=var_constraint,
    )
    expected_combined_constraint = "ConstraintCombination(AttributeConstraint({'STASH': 'm01s03i236'}), Constraint(name='test'), <built-in function and_>)"
    assert repr(combined_constraint) == expected_combined_constraint
    var_constraint_2 = constraints.generate_var_constraint("test_2")
    combined_constraint = constraints.combine_constraints(
        stash_constraint,
        a=var_constraint,
        b=var_constraint_2,
    )
    expected_combined_constraint = "ConstraintCombination(ConstraintCombination(AttributeConstraint({'STASH': 'm01s03i236'}), Constraint(name='test'), <built-in function and_>), Constraint(name='test_2'), <built-in function and_>)"
    assert repr(combined_constraint) == expected_combined_constraint



#! /usr/bin/env python3
# © Crown copyright, Met Office (2022-2025) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates the rose-meta.conf file from a template.

To use, simply execute this script and it will regenerate the rose-meta.conf
from its template.

$ ./scripts/generate_rose_meta.py
"""

from pathlib import Path

import jinja2

rose_meta_folder = Path(__file__).parent.parent / "cset-workflow/meta/"
source = rose_meta_folder / "rose-meta.conf.jinja2"
destination = rose_meta_folder / "rose-meta.conf"

env = jinja2.Environment(autoescape=False, lstrip_blocks=True, trim_blocks=True)
template = env.from_string(source.read_text())
destination.write_text(template.render())



# © Crown copyright, Met Office (2022-2025) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operators to perform various kind of collapse on either 1 or 2 dimensions."""

import warnings

import iris
import iris.analysis
import iris.coord_categorisation
import iris.cube
import iris.util

from CSET._common import iter_maybe
from CSET.operators._utils import is_time_aggregatable
from CSET.operators.aggregate import (
    add_hour_coordinate,
    ensure_aggregatable_across_cases,
)


def collapse(
    cubes: iris.cube.Cube | iris.cube.CubeList,
    coordinate: str | list[str],
    method: str,
    additional_percent: float = None,
    **kwargs,
) -> iris.cube.Cube | iris.cube.CubeList:
    """Collapse coordinate(s) of a single cube or of every cube in a cube list.

    Collapses similar fields in each cube into a cube collapsing around the
    specified coordinate(s) and method. This could be a (weighted) mean or
    percentile.

    Arguments
    ---------
    cubes: iris.cube.Cube | iris.cube.CubeList
         Cube or CubeList to collapse and iterate over one dimension
    coordinate: str | list[str]
         Coordinate(s) to collapse over e.g. 'time', 'longitude', 'latitude',
         'model_level_number', 'realization'. A list of multiple coordinates can
         be given.
    method: str
         Type of collapse i.e. method: 'MEAN', 'MAX', 'MIN', 'MEDIAN',
         'PERCENTILE' getattr creates iris.analysis.MEAN, etc For PERCENTILE
         YAML file requires i.e. method: 'PERCENTILE' additional_percent: 90

    Returns
    -------
    collapsed_cubes: iris.cube.Cube | iris.cube.CubeList
        Single variable but several methods of aggregation

    Raises
    ------
    ValueError
        If additional_percent wasn't supplied while using PERCENTILE method.
    """
    if method == "PERCENTILE" and additional_percent is None:
        raise ValueError("Must specify additional_percent")
    collapsed_cubes = iris.cube.CubeList([])
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", "Cannot check if coordinate is contiguous", UserWarning
        )
        warnings.filterwarnings(
            "ignore", "Collapsing spatial coordinate.+without weighting", UserWarning
        )
        for cube in iter_maybe(cubes):
            if method == "PERCENTILE":
                collapsed_cubes.append(
                    cube.collapsed(
                        coordinate,
                        getattr(iris.analysis, method),
                        percent=additional_percent,
                    )
                )
            else:
                collapsed_cubes.append(
                    cube.collapsed(coordinate, getattr(iris.analysis, method))
                )
    if len(collapsed_cubes) == 1:
        return collapsed_cubes[0]
    else:
        return collapsed_cubes


def collapse_by_lead_time(
    cube: iris.cube.Cube | iris.cube.CubeList,
    method: str,
    additional_percent: float = None,
    **kwargs,
) -> iris.cube.Cube:
    """Collapse a cube around lead time for multiple cases.

    First checks if the data can be aggregated by lead time easily. Then
    collapses by lead time for a specified method using the collapse function.
    This operator provides the average of all T+1, T+2, etc.

    Arguments
    ---------
    cube: iris.cube.Cube | iris.cube.CubeList
        Cube to collapse by lead time or CubeList that will be converted
        to a cube before collapsing by lead time.
    method: str
         Type of collapse i.e. method: 'MEAN', 'MAX', 'MIN', 'MEDIAN',
         'PERCENTILE'. For 'PERCENTILE' the additional_percent must be specified.

    Returns
    -------
    cube: iris.cube.Cube
        Single variable collapsed by lead time based on chosen method.

    Raises
    ------
    ValueError
        If additional_percent wasn't supplied while using PERCENTILE method.
    """
    if method == "PERCENTILE" and additional_percent is None:
        raise ValueError("Must specify additional_percent")
    # Ensure the cube can be aggregated over multiple cases.
    cube_to_collapse = ensure_aggregatable_across_cases(cube)
    # Collapse by lead time.
    if method == "PERCENTILE":
        collapsed_cube = collapse(
            cube_to_collapse,
            "forecast_reference_time",
            method,
            additional_percent=additional_percent,
        )
    else:
        collapsed_cube = collapse(cube_to_collapse, "forecast_reference_time", method)
    return collapsed_cube


def collapse_by_hour_of_day(
    cube: iris.cube.Cube | iris.cube.CubeList,
    method: str,
    additional_percent: float = None,
    multi_case: bool = True,
    **kwargs,
) -> iris.cube.Cube:
    """Collapse a cube by hour of the day.

    Collapses a cube by hour of the day in the time coordinates provided by the
    model. It is useful for creating diurnal cycle plots. It aggregates all
    00 UTC together regardless of lead time.

    Arguments
    ---------
    cube: iris.cube.Cube | iris.cube.CubeList
        Cube to collapse and iterate over one dimension or CubeList to
        convert to a cube and then collapse prior to aggregating by hour.
        If a CubeList is provided multi_case must be set to True as the Cube List
        should only contain cubes of multiple dates and not different variables
        or models. A cube that only contains one time dimension must have
        multi_case set to False as it contains only one forecast. A cube
        containing two time dimensions, e.g., 'forecast_reference_time' and
        'forecast_period' must have multi_case set to True as it will contain
        multiple forecasts.
    method: str
        Type of collapse i.e. method: 'MEAN', 'MAX', 'MIN', 'MEDIAN',
        'PERCENTILE'. For 'PERCENTILE' the additional_percent must be specified.
    multi_case: boolean, optional
        Default is True. If True multiple cases will be aggregated by hour of
        day; if False a single forecast will be aggregated by hour of day.
        Information around the usage of multi_case is provided above under the
        description for the cube argument. It is kept as an argument rather
        than being automatically generated to maintain traceability for the
        users actions.

    Returns
    -------
    cube: iris.cube.Cube
        Single variable but several methods of aggregation.

    Raises
    ------
    ValueError
        If additional_percent wasn't supplied while using PERCENTILE method.
    TypeError
        If a CubeList is given and multi_case is not True;
        if a Cube is given and contains two time dimensions and multi_case is not True;
        if a Cube is given and contains one time dimensions and multi_case is not False.

    Notes
    -----
    Collapsing of the cube is around the 'time' coordinate. The coordinates are
    first grouped by the hour of day, and then aggregated by the hour of day
    to create a diurnal cycle. This operator is applicable for both single
    forecasts and for multiple forecasts. The hour used is based on the units of
    the time coordinate. If the time coordinate is in UTC, hour will be in UTC.

    To apply this operator successfully there must only be one time dimension.
    Should a MultiDim exception be raised the user first needs to apply the
    collapse operator to reduce the time dimensions before applying this
    operator. If multi_case is true the collapse_by_lead_time operator is
    applied and performs this step.
    """
    if method == "PERCENTILE" and additional_percent is None:
        raise ValueError("Must specify additional_percent")
    elif (
        isinstance(cube, iris.cube.Cube)
        and is_time_aggregatable(cube)
        and not multi_case
    ):
        raise TypeError(
            "multi_case must be true for a cube containing two time dimensions"
        )
    elif (
        isinstance(cube, iris.cube.Cube)
        and not is_time_aggregatable(cube)
        and multi_case
    ):
        raise TypeError(
            "multi_case must be false for a cube containing one time dimension"
        )
    elif isinstance(cube, iris.cube.CubeList) and not multi_case:
        raise TypeError("multi_case must be true for a CubeList")

    if multi_case:
        # Collapse by lead time to get a single time dimension.
        cube = collapse_by_lead_time(
            cube, method, additional_percent=additional_percent
        )

    # Categorise the time coordinate by hour of the day.
    cube = add_hour_coordinate(cube)
    # Aggregate by the new category coordinate.
    if method == "PERCENTILE":
        collapsed_cube = cube.aggregated_by(
            "hour", getattr(iris.analysis, method), percent=additional_percent
        )
    else:
        collapsed_cube = cube.aggregated_by("hour", getattr(iris.analysis, method))

    # Remove unnecessary time coordinates.
    collapsed_cube.remove_coord("time")
    collapsed_cube.remove_coord("forecast_period")

    # Sort hour coordinate.
    collapsed_cube.coord("hour").points.sort()

    # Remove forecast_reference_time if a single case, as collapse_by_lead_time
    # will have effectively done this if multi_case is True.
    if not multi_case:
        collapsed_cube.remove_coord("forecast_reference_time")

    # Promote "hour" to dim_coord.
    iris.util.promote_aux_coord_to_dim_coord(collapsed_cube, "hour")
    return collapsed_cube


def collapse_by_validity_time(
    cube: iris.cube.Cube | iris.cube.CubeList,
    method: str,
    additional_percent: float = None,
    **kwargs,
) -> iris.cube.Cube:
    """Collapse a cube around validity time for multiple cases.

    First checks if the data can be aggregated easily. Then creates a new cube
    by slicing over the time dimensions, removing the time dimensions,
    re-merging the data, and creating a new time coordinate. It then collapses
    by the new time coordinate for a specified method using the collapse
    function.

    Arguments
    ---------
    cube: iris.cube.Cube | iris.cube.CubeList
        Cube to collapse by validity time or CubeList that will be converted
        to a cube before collapsing by validity time.
    method: str
         Type of collapse i.e. method: 'MEAN', 'MAX', 'MIN', 'MEDIAN',
         'PERCENTILE'. For 'PERCENTILE' the additional_percent must be specified.

    Returns
    -------
    cube: iris.cube.Cube
        Single variable collapsed by lead time based on chosen method.

    Raises
    ------
    ValueError
        If additional_percent wasn't supplied while using PERCENTILE method.
    """
    if method == "PERCENTILE" and additional_percent is None:
        raise ValueError("Must specify additional_percent")
    # Ensure the cube can be aggregated over multiple times.
    cube_to_collapse = ensure_aggregatable_across_cases(cube)
    # Convert to a cube that is split by validity time.
    # Slice over cube by both time dimensions to create a CubeList.
    new_cubelist = iris.cube.CubeList(
        cube_to_collapse.slices_over(["forecast_period", "forecast_reference_time"])
    )
    # Remove forecast_period and forecast_reference_time coordinates.
    for sub_cube in new_cubelist:
        sub_cube.remove_coord("forecast_period")
        sub_cube.remove_coord("forecast_reference_time")
    # Create new CubeList by merging with unique = False to produce a validity
    # time cube.
    merged_list_1 = new_cubelist.merge(unique=False)
    # Create a new "fake" coordinate and apply to each remaining cube to allow
    # final merging to take place into a single cube.
    equalised_validity_time = iris.coords.AuxCoord(
        points=0, long_name="equalised_validity_time", units="1"
    )
    for sub_cube, eq_valid_time in zip(
        merged_list_1, range(len(merged_list_1)), strict=True
    ):
        sub_cube.add_aux_coord(equalised_validity_time.copy(points=eq_valid_time))

    # Merge CubeList to create final cube.
    final_cube = merged_list_1.merge_cube()
    # Collapse over fake_time_coord to represent collapsing over validity time.
    if method == "PERCENTILE":
        collapsed_cube = collapse(
            final_cube,
            "equalised_validity_time",
            method,
            additional_percent=additional_percent,
        )
    else:
        collapsed_cube = collapse(final_cube, "equalised_validity_time", method)
    return collapsed_cube


# TODO
# Collapse function that calculates means, medians etc across members of an
# ensemble or stratified groups. Need to allow collapse over realisation
# dimension for fixed time. Hence will require reading in of CubeList



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for send_email workflow utility."""

import subprocess

import pytest

from CSET._workflow_utils import send_email


def test_get_home_page_address_public_html(monkeypatch):
    """Test public_html path gets parsed properly."""
    monkeypatch.setenv("WEB_DIR", "/home/user/public_html/CSET")
    monkeypatch.setenv("WEB_ADDR", "https://example.com/~user")
    expected = "https://example.com/~user/CSET"
    actual = send_email.get_home_page_address()
    assert actual == expected


def test_get_home_page_address_Public(monkeypatch):
    """Test Public path and trailing slash on WEB_ADDR."""
    monkeypatch.setenv("WEB_DIR", "/home/user/Public/CSET")
    monkeypatch.setenv("WEB_ADDR", "https://example.com/~user/")
    expected = "https://example.com/~user/CSET"
    actual = send_email.get_home_page_address()
    assert actual == expected


def test_get_home_page_address_no_known_web_directory(monkeypatch):
    """Exception when web directory cannot be determined."""
    monkeypatch.setenv("WEB_DIR", "/home/user/CSET")
    monkeypatch.setenv("WEB_ADDR", "https://example.com/~user")
    with pytest.raises(ValueError, match="Cannot determine web address."):
        send_email.get_home_page_address()


def test_get_home_page_address_no_WEB_DIR(monkeypatch):
    """Exception when WEB_DIR not set."""
    monkeypatch.setenv("WEB_ADDR", "https://example.com/~user")
    with pytest.raises(ValueError, match="WEB_DIR not set."):
        send_email.get_home_page_address()


def test_get_home_page_address_no_WEB_ADDR(monkeypatch):
    """Exception when WEB_ADDR not set."""
    monkeypatch.setenv("WEB_DIR", "/home/user/Public/CSET")
    with pytest.raises(ValueError, match="WEB_ADDR not set."):
        send_email.get_home_page_address()


def test_send_email(monkeypatch):
    """Test send_email run."""
    monkeypatch.setenv("WEB_DIR", "/home/user/public_html/CSET")
    monkeypatch.setenv("WEB_ADDR", "https://example.com/~user")

    def mock_subprocess_run(args, check, shell):
        assert (
            args
            == 'printf "The webpage for your run of CSET is now ready. You can view it here:\nhttps://example.com/~user/CSET" | mail -s "CSET webpage ready" -S "from=notifications" "$USER"'
        )
        assert check
        assert shell

    monkeypatch.setattr(subprocess, "run", mock_subprocess_run)
    send_email.run()


def test_send_email_invalid_environment_variables(monkeypatch):
    """Test send_email run."""

    def mock_subprocess_run(args, check, shell):
        assert (
            args
            == 'printf "The webpage for your run of CSET is now ready, though the address could not be determined.\nCheck that WEB_ADDR and WEB_DIR are set correctly, then consider filing a bug report at https://github.com/MetOffice/CSET" | mail -s "CSET webpage ready" -S "from=notifications" "$USER"'
        )
        assert check
        assert shell

    monkeypatch.setattr(subprocess, "run", mock_subprocess_run)
    send_email.run()



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test filter operators."""

import iris.cube
import numpy as np
import pytest

from CSET.operators import constraints, filters, read

constraint_single = constraints.combine_constraints(
    constraints.generate_stash_constraint("m01s03i236"),
    a=constraints.generate_cell_methods_constraint([]),
)


def test_filter_cubes_cubelist(cubes):
    """Test filtering a CubeList."""
    constraint = constraints.generate_cell_methods_constraint([])
    cube = filters.filter_cubes(cubes, constraint)
    assert cube.cell_methods == ()
    expected_cube = "<iris 'Cube' of air_temperature / (K) (time: 3; grid_latitude: 17; grid_longitude: 13)>"
    assert repr(cube) == expected_cube


def test_filter_cubes_cube(cube):
    """Test filtering a Cube."""
    constraint = constraints.generate_cell_methods_constraint([])
    single_cube = filters.filter_cubes(cube, constraint)
    assert repr(cube) == repr(single_cube)


def test_filter_cubes_multiple_returned_exception(cubes):
    """Test for exception when multiple cubes returned."""
    constraint = constraints.generate_stash_constraint("m01s03i236")
    with pytest.raises(ValueError):
        filters.filter_cubes(cubes, constraint)


def test_filter_cubes_no_level_coord(cube):
    """Test a cube without the level coordinate is passed through."""
    pressure_constraint = constraints.generate_level_constraint(
        coordinate="pressure", levels=[]
    )
    assert filters.filter_cubes(cube, pressure_constraint)


def test_filter_cubes_no_level_coord_none_returned(cube):
    """Test exception when pressure coordinate is excluded."""
    pressure_constraint = constraints.generate_level_constraint(
        coordinate="pressure", levels=[]
    )
    cube.add_aux_coord(iris.coords.DimCoord(100, var_name="pressure"))
    with pytest.raises(ValueError):
        filters.filter_cubes(cube, pressure_constraint)


def test_filter_multiple_cubes(cubes):
    """Test to a CubeList of a single Cube."""
    filtered_cubes = filters.filter_multiple_cubes(cubes, c=constraint_single)
    assert isinstance(filtered_cubes, iris.cube.CubeList)
    assert len(filtered_cubes) == 1
    assert filtered_cubes[0].cell_methods == ()


def test_filter_multiple_cubes_single_cube(cube):
    """Test filtering a Cube."""
    filtered_cube = filters.filter_multiple_cubes(cube, c=constraint_single)
    assert isinstance(filtered_cube, iris.cube.CubeList)
    assert len(filtered_cube) == 1


def test_filter_multiple_cubes_multiple_out(cubes):
    """Test filtering for multiple Cubes."""
    constraint_single_2 = constraints.combine_constraints(
        constraints.generate_stash_constraint("m01s03i236"),
        a=constraints.generate_cell_methods_constraint(
            [
                iris.coords.CellMethod(
                    method="minimum", coords="time", intervals="1 hour"
                )
            ]
        ),
    )
    filtered_multi_cubes = filters.filter_multiple_cubes(
        cubes, c1=constraint_single, c2=constraint_single_2
    )
    assert len(filtered_multi_cubes) == 2


def test_filter_multiple_cubes_no_constraint_exception(cubes):
    """Test exception when no constraints given."""
    with pytest.raises(ValueError):
        filters.filter_multiple_cubes(cubes)


def test_filter_multiple_cubes_returned(cubes):
    """Test exception when multiple cubes returned."""
    constraint_multiple = constraints.generate_stash_constraint("m01s03i236")
    with pytest.raises(ValueError):
        filters.filter_multiple_cubes(cubes, c=constraint_multiple)


def test_filter_multiple_cubes_none_returned(cubes):
    """Test exception when no Cubes returned."""
    constraint_none = constraints.generate_stash_constraint("m01s01i001")
    with pytest.raises(ValueError):
        filters.filter_multiple_cubes(cubes, c=constraint_none)


# Session scope fixtures, so the test data only has to be loaded once.
@pytest.fixture(scope="session")
def load_vertical_coord_cubelist() -> iris.cube.CubeList:
    """Get a cubelist with multiple vertical level cubes."""
    return read.read_cubes("tests/test_data/vertlevtestdata.nc", "y_wind")


def test_generate_level_constraint_return_single_level(load_vertical_coord_cubelist):
    """For a cubelist that contains 3 cubes on different vertical levels.

    Return one without a vertical coordinate.
    """
    constraint_1 = constraints.generate_level_constraint(
        coordinate="pressure", levels=[]
    )
    constraint_2 = constraints.generate_level_constraint(
        coordinate="model_level_number", levels=[]
    )
    combined = constraints.combine_constraints(constraint_1, a=constraint_2)
    cube = load_vertical_coord_cubelist.extract_cube(combined)
    coords = [coord.name() for coord in cube.coords()]
    assert "pressure" not in coords
    assert "model_level_number" not in coords


def test_generate_level_constraint_return_all_pressure(load_vertical_coord_cubelist):
    """For a cubelist that contains 3 cubes on different vertical levels.

    Return one with a pressure on all levels.
    """
    constraint = constraints.generate_level_constraint(
        coordinate="pressure", levels="*"
    )
    coord = load_vertical_coord_cubelist.extract_cube(constraint).coord("pressure")
    assert coord.shape == (34,)


def test_generate_level_constraint_return_all_model_levels(
    load_vertical_coord_cubelist,
):
    """For a cubelist that contains 3 cubes on different vertical levels.

    Return one with a model level on all levels.
    """
    constraint = constraints.generate_level_constraint(
        coordinate="model_level_number", levels="*"
    )
    coord = load_vertical_coord_cubelist.extract_cube(constraint).coord(
        "model_level_number"
    )
    assert coord.shape == (70,)


def test_generate_mask_fail_wrong_condition(cube):
    """Fails for a wrong condition."""
    with pytest.raises(ValueError):
        filters.generate_mask(cube, "!<", 10)


def test_generate_mask_equal_to(cube):
    """Generates a mask with values equal to a specified value."""
    mask = cube.copy()
    mask.data = np.zeros(mask.data.shape)
    mask.data[cube.data == 276] = 1
    assert np.allclose(
        filters.generate_mask(cube, "==", 276).data,
        mask.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_generate_mask_not_equal_to(cube):
    """Generates a mask with values not equal to a specified value."""
    mask = cube.copy()
    mask.data = np.zeros(mask.data.shape)
    mask.data[cube.data != 276] = 1
    assert np.allclose(
        filters.generate_mask(cube, "!=", 276).data,
        mask.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_generate_mask_greater_than(cube):
    """Generates a mask with values greater than a specified value."""
    mask = cube.copy()
    mask.data = np.zeros(mask.data.shape)
    mask.data[cube.data > 276] = 1
    assert np.allclose(
        filters.generate_mask(cube, ">", 276).data,
        mask.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_generate_mask_greater_equal_to(cube):
    """Generates a mask with values greater than or equal to a specified value."""
    mask = cube.copy()
    mask.data = np.zeros(mask.data.shape)
    mask.data[cube.data >= 276] = 1
    assert np.allclose(
        filters.generate_mask(cube, ">=", 276).data,
        mask.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_generate_mask_less_than(cube):
    """Generates a mask with values less than a specified value."""
    mask = cube.copy()
    mask.data = np.zeros(mask.data.shape)
    mask.data[cube.data < 276] = 1
    assert np.allclose(
        filters.generate_mask(cube, "<", 276).data,
        mask.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_generate_mask_less_equal_to(cube):
    """Generates a mask with values less than or equal to a specified value."""
    mask = cube.copy()
    mask.data = np.zeros(mask.data.shape)
    mask.data[cube.data <= 276] = 1
    assert np.allclose(
        filters.generate_mask(cube, "<=", 276).data,
        mask.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_apply_mask(cube):
    """Apply a mask to a cube."""
    mask = filters.generate_mask(cube, "==", 276)
    mask.data[mask.data == 0] = np.nan
    mask.data[~np.isnan(mask.data)] = 1
    test_data = cube.copy()
    test_data.data *= mask.data
    assert np.allclose(
        filters.apply_mask(cube, mask).data,
        test_data.data,
        rtol=1e-06,
        atol=1e-02,
        equal_nan=True,
    )



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Operations on recipes."""

import importlib.resources
import logging
import sys
import warnings
from collections.abc import Iterable
from pathlib import Path

import ruamel.yaml


class FileExistsWarning(UserWarning):
    """Warning a file already exists, and some unusual action shall be taken."""


def _version_agnostic_importlib_resources_file() -> Path:
    """Transitional wrapper to importlib.resources.files().

    Importlib behaviour changed in 3.12 to avoid circular dependencies.
    """
    if sys.version_info.minor >= 12:
        input_dir = importlib.resources.files()
    else:
        import CSET.recipes

        input_dir = importlib.resources.files(CSET.recipes)
    return input_dir


def _recipe_files_in_tree(
    recipe_name: str = None, input_dir: Path = None
) -> Iterable[Path]:
    """Yield recipe file Paths matching the recipe name."""
    if recipe_name is None:
        recipe_name = ""
    if input_dir is None:
        input_dir = _version_agnostic_importlib_resources_file()
    for file in input_dir.iterdir():
        logging.debug("Testing %s", file)
        if recipe_name in file.name and file.is_file() and file.suffix == ".yaml":
            yield file
        elif file.is_dir() and file.name[0] != "_":  # Excludes __pycache__
            yield from _recipe_files_in_tree(recipe_name, file)


def _get_recipe_file(recipe_name: str, input_dir: Path = None) -> Path:
    """Return a Path to the recipe file."""
    if input_dir is None:
        input_dir = _version_agnostic_importlib_resources_file()
    file = input_dir / recipe_name
    logging.debug("Getting recipe: %s", file)
    if not file.is_file():
        raise FileNotFoundError("Recipe file does not exist.", recipe_name)
    return file


def unpack_recipe(recipe_dir: Path, recipe_name: str) -> None:
    """
    Unpacks recipes files into a directory, creating it if it doesn't exist.

    Parameters
    ----------
    recipe_dir: Path
        Path to a directory into which to unpack the recipe files.
    recipe_name: str
        Name of recipe to unpack.

    Raises
    ------
    FileExistsError
        If recipe_dir already exists, and is not a directory.

    OSError
        If recipe_dir cannot be created, such as insufficient permissions, or
        lack of space.
    """
    file = _get_recipe_file(recipe_name)
    recipe_dir.mkdir(parents=True, exist_ok=True)
    output_file = recipe_dir / file.name
    logging.debug("Saving recipe to %s", output_file)
    if output_file.exists():
        warnings.warn(
            f"{file.name} already exists in target directory, skipping.",
            FileExistsWarning,
            stacklevel=2,
        )
        return
    logging.info("Unpacking %s to %s", file.name, output_file)
    output_file.write_bytes(file.read_bytes())


def list_available_recipes() -> None:
    """List available recipes to stdout."""
    print("Available recipes:")
    for file in _recipe_files_in_tree():
        print(f"\t{file.name}")


def detail_recipe(recipe_name: str) -> None:
    """Detail the recipe to stdout.

    If multiple recipes match the given name they will all be displayed.

    Parameters
    ----------
    recipe_name: str
        Partial match for the recipe name.
    """
    for file in _recipe_files_in_tree(recipe_name):
        with ruamel.yaml.YAML(typ="safe", pure=True) as yaml:
            recipe = yaml.load(file)
        print(f"\n\t{file.name}\n\t{''.join('─' * len(file.name))}\n")
        print(recipe.get("description"))



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for fetch_data workflow utility."""

import datetime
import hashlib
from pathlib import Path

import pytest

from CSET._workflow_utils import fetch_data


def mock_get_needed_environment_variables():
    """Minimal set of environment variables for running a mock fetch_data."""
    return {
        "data_period": None,
        "data_time": None,
        "date_type": None,
        "forecast_length": None,
        "forecast_offset": None,
        "model_identifier": "1",
        "raw_path": None,
        "rose_datac": "/tmp/cycle/20000101T0000Z",
    }


def mock_template_file_path(*args, **kwargs):
    """List of paths for testing with."""
    return [f"path_{n}" for n in range(5)]


def test_get_needed_environment_variables(monkeypatch):
    """Needed environment variables are loaded."""
    duration_raw = "PT1H"
    duration = datetime.timedelta(hours=1)
    date_raw = "20000101T0000Z"
    date = datetime.datetime(2000, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    path = "/path/to/data"
    number_raw = "1"

    monkeypatch.setenv("ANALYSIS_OFFSET", duration_raw)
    monkeypatch.setenv("ANALYSIS_LENGTH", duration_raw)
    monkeypatch.setenv("CYLC_TASK_CYCLE_POINT", date_raw)
    monkeypatch.setenv("DATA_PATH", path)
    monkeypatch.setenv("DATA_PERIOD", duration_raw)
    monkeypatch.setenv("MODEL_IDENTIFIER", number_raw)
    monkeypatch.setenv("ROSE_DATAC", path)
    monkeypatch.setenv("DATE_TYPE", "validity")

    expected = {
        "data_period": duration,
        "data_time": date,
        "date_type": "validity",
        "forecast_length": duration,
        "forecast_offset": duration,
        "model_identifier": number_raw,
        "raw_path": path,
        "rose_datac": path,
    }
    actual = fetch_data._get_needed_environment_variables()
    assert actual == expected, "Unexpected values from reading environment variables"


def test_get_needed_environment_variables_data_period_handling(monkeypatch):
    """Handle data_period dependent on date type."""
    duration_raw = "PT1H"
    date_raw = "20000101T0000Z"
    path = "/path/to/data"
    number_raw = "1"

    monkeypatch.setenv("ANALYSIS_OFFSET", duration_raw)
    monkeypatch.setenv("ANALYSIS_LENGTH", duration_raw)
    monkeypatch.setenv("CYLC_TASK_CYCLE_POINT", date_raw)
    monkeypatch.setenv("DATA_PATH", path)
    monkeypatch.setenv("MODEL_IDENTIFIER", number_raw)
    monkeypatch.setenv("ROSE_DATAC", path)

    # Check DATA_PERIOD is not there for initiation.
    monkeypatch.setenv("DATE_TYPE", "initiation")
    initiation_actual = fetch_data._get_needed_environment_variables()
    assert initiation_actual["data_period"] is None, (
        "data_period should not be set for initiation time"
    )

    # Check exception when data period is not specified for validity time.
    monkeypatch.setenv("DATE_TYPE", "validity")
    with pytest.raises(KeyError):
        fetch_data._get_needed_environment_variables()


def test_fetch_data(monkeypatch):
    """Check get_file is called appropriately when fetching data."""
    actually_called = False

    class MockFileRetriever(fetch_data.FileRetrieverABC):
        def get_file(self, file_path: str, output_dir: str) -> None:
            nonlocal actually_called
            actually_called = True
            return True

    monkeypatch.setattr(
        fetch_data,
        "_get_needed_environment_variables",
        mock_get_needed_environment_variables,
    )
    monkeypatch.setattr(fetch_data, "_template_file_path", mock_template_file_path)
    fetch_data.fetch_data(MockFileRetriever)
    assert actually_called


def test_fetch_data_no_files_found(monkeypatch):
    """Check exception is raised when no files found."""
    actually_called = False

    class MockFileRetrieverNoFiles(fetch_data.FileRetrieverABC):
        def get_file(self, file_path: str, output_dir: str) -> None:
            nonlocal actually_called
            actually_called = True
            return False

    monkeypatch.setattr(
        fetch_data,
        "_get_needed_environment_variables",
        mock_get_needed_environment_variables,
    )
    monkeypatch.setattr(fetch_data, "_template_file_path", mock_template_file_path)
    with pytest.raises(FileNotFoundError, match="No files found for model!"):
        fetch_data.fetch_data(MockFileRetrieverNoFiles)

    assert actually_called


def test_template_file_path_validity_time():
    """Test filling path placeholders for validity time."""
    actual = fetch_data._template_file_path(
        "/path/%Y-%m-%d.nc",
        "validity",
        datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc),
        datetime.timedelta(days=5),
        datetime.timedelta(),
        datetime.timedelta(days=1),
    )
    expected = [
        "/path/2000-01-01.nc",
        "/path/2000-01-02.nc",
        "/path/2000-01-03.nc",
        "/path/2000-01-04.nc",
        "/path/2000-01-05.nc",
    ]
    assert actual == expected


def test_template_file_path_initiation_time():
    """Test filling path placeholders for initiation time."""
    actual = fetch_data._template_file_path(
        "/path/%Y-%m-%d.nc",
        "initiation",
        datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc),
        datetime.timedelta(days=5),
        datetime.timedelta(),
        datetime.timedelta(days=1),
    )
    expected = ["/path/2000-01-01.nc"]
    assert actual == expected


def test_template_file_path_lead_time():
    """Test filling path placeholders for lead time."""
    actual = fetch_data._template_file_path(
        "/path/%N.nc",
        "initiation",
        datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc),
        datetime.timedelta(hours=5, seconds=1),
        datetime.timedelta(hours=1),
        datetime.timedelta(hours=1),
    )
    expected = [
        "/path/001.nc",
        "/path/002.nc",
        "/path/003.nc",
        "/path/004.nc",
        "/path/005.nc",
    ]
    assert actual == expected


def test_template_file_path_invalid_date_type():
    """Test error on invalid date type."""
    with pytest.raises(ValueError, match="Invalid date type:"):
        fetch_data._template_file_path(None, "Other", None, None, None, None)


def test_FilesystemFileRetriever(tmp_path):
    """Test retrieving a file from the filesystem."""
    with fetch_data.FilesystemFileRetriever() as ffr:
        files_found = ffr.get_file("tests/test_data/exeter_em*.nc", str(tmp_path))
    assert (tmp_path / "exeter_em01.nc").is_file()
    assert (tmp_path / "exeter_em02.nc").is_file()
    assert files_found


def test_FilesystemFileRetriever_no_files(tmp_path, caplog):
    """Test warning when no files match the requested path."""
    with fetch_data.FilesystemFileRetriever() as ffr:
        # Should warn, but not error.
        files_found = ffr.get_file("/non-existent/file.nc", str(tmp_path))
    log_record = next(
        filter(
            lambda record: record.message.startswith(
                "file_path does not match any files:"
            ),
            caplog.records,
        ),
        None,
    )
    assert log_record
    assert log_record.levelname == "WARNING"
    assert not files_found


def test_FilesystemFileRetriever_copy_error(caplog):
    """Test warning when file copy errors."""
    with fetch_data.FilesystemFileRetriever() as ffr:
        # Please don't run as root.
        files_found = ffr.get_file("tests/test_data/air_temp.nc", "/usr/bin")
    assert not Path("/usr/bin/air_temp.nc").is_file()
    log_record = next(
        filter(
            lambda record: record.message.startswith("Failed to copy"), caplog.records
        ),
        None,
    )
    assert log_record
    assert log_record.levelname == "WARNING"
    assert not files_found


@pytest.mark.network
def test_HTTPFileRetriever(tmp_path):
    """Test retrieving a file via HTTP."""
    url = "https://github.com/MetOffice/CSET/raw/48dc1d29846604aacb8d370b82bca31405931c87/tests/test_data/exeter_em01.nc"
    with fetch_data.HTTPFileRetriever() as hfr:
        files_found = hfr.get_file(url, str(tmp_path))
    file = tmp_path / "exeter_em01.nc"
    assert file.is_file()
    # Check file hash is correct, indicating a non-corrupt download.
    expected_hash = "67899970eeca75b9378f0275ce86db3d1d613f2bc7a178540912848dc8a69ca7"
    actual_hash = hashlib.sha256(file.read_bytes()).hexdigest()
    assert actual_hash == expected_hash
    assert files_found


@pytest.mark.network
def test_HTTPFileRetriever_no_files(tmp_path, caplog):
    """Test warning rather than error when requested URL does not exist."""
    with fetch_data.HTTPFileRetriever() as hfr:
        # Should warn, but not error.
        files_found = hfr.get_file("http://httpbin.org/status/404", str(tmp_path))
    log_record = next(
        filter(
            lambda record: record.message.startswith(
                "Failed to retrieve http://httpbin.org/status/404, error:"
            ),
            caplog.records,
        ),
        None,
    )
    assert log_record
    assert log_record.levelname == "WARNING"
    assert not files_found



# © Crown copyright, Met Office (2022-2024) and CSET contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test aggregate operators."""

import iris
import iris.cube
import numpy as np
import pytest

from CSET.operators import aggregate


def test_aggregate(cube):
    """Aggregate time to 2 hour intervals."""
    # Set test interval to 2 hours.
    interval = "PT2H"

    # Introduces new coordinate for cube based on existing
    # coordinate which has equal or less increments.

    # Test adding further coordinate.
    aggregated_cube = aggregate.time_aggregate(
        cube.copy(), method="SUM", interval_iso=interval
    )

    # Check if number of coords on aggregated cube is same as original cube.
    assert len(aggregated_cube.coords()) == len(cube.coords()), (
        "aggregated cube does not have additional aux coordinate"
    )


def test_ensure_aggregatable_across_cases_true_aggregatable_cube(
    long_forecast_multi_day,
):
    """Check that an aggregatable cube is returned with no changes."""
    assert np.allclose(
        aggregate.ensure_aggregatable_across_cases(long_forecast_multi_day).data,
        long_forecast_multi_day.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_ensure_aggregatable_across_cases_false_aggregatable_cube(long_forecast):
    """Check that a non-aggregatable cube raises an error."""
    with pytest.raises(ValueError):
        aggregate.ensure_aggregatable_across_cases(long_forecast)


def test_ensure_aggregatable_across_cases_cubelist(
    long_forecast_many_cubes, long_forecast_multi_day
):
    """Check that a CubeList turns into an aggregatable Cube."""
    # Check output is a Cube.
    output_data = aggregate.ensure_aggregatable_across_cases(long_forecast_many_cubes)
    assert isinstance(output_data, iris.cube.Cube)
    # Check output can be aggregated in time.
    assert isinstance(
        aggregate.ensure_aggregatable_across_cases(output_data), iris.cube.Cube
    )
    # Check output is identical to a pre-calculated cube.
    pre_calculated_data = long_forecast_multi_day
    assert np.allclose(
        pre_calculated_data.data,
        output_data.data,
        rtol=1e-06,
        atol=1e-02,
    )


def test_add_hour_coordinate(long_forecast):
    """Check that a Cube has an hour coordinate added to it."""
    cube_with_hour_coordinate = aggregate.add_hour_coordinate(long_forecast)
    hour_coord = cube_with_hour_coordinate.coord("hour")
    assert hour_coord
    assert hour_coord.units == "hours"
    assert len(set(hour_coord.points)) == 24


def test_add_hour_coordinate_cubelist(long_forecast_many_cubes):
    """Check that a CubeList has hour coordinates added to it."""
    cubes_with_hour_coordinate = aggregate.add_hour_coordinate(long_forecast_many_cubes)
    assert isinstance(cubes_with_hour_coordinate, iris.cube.CubeList)
    for cube in cubes_with_hour_coordinate:
        hour_coord = cube.coord("hour")
        assert hour_coord
        assert hour_coord.units == "hours"
        assert len(set(hour_coord.points)) == 24
